@@ -3,18 +3,17 @@
   array.c -
 
   $Author: matz $
-  $Date: 1996/12/25 10:42:18 $
+  $Date: 1998/11/09 09:11:48 $
   created at: Fri Aug  6 09:46:12 JST 1993
 
-  Copyright (C) 1993-1996 Yukihiro Matsumoto
+  Copyright (C) 1993-1998 Yukihiro Matsumoto
 
 ************************************************/
 
 #include "ruby.h"
 
-VALUE cArray;
 
-VALUE rb_to_a();
+VALUE cArray;
 
 #define ARY_DEFAULT_SIZE 16
 
@@ -28,6 +27,17 @@ memclear(mem, size)
     }
 }
 
+static void
+memfill(mem, size, val)
+    register VALUE *mem;
+    register int size;
+    register VALUE val;
+{
+    while (size--) {
+	*mem++ = val;
+    }
+}
+
 #define ARY_FREEZE   FL_USER1
 
 static void
@@ -64,14 +74,17 @@ ary_new2(len)
     NEWOBJ(ary, struct RArray);
     OBJSETUP(ary, cArray, T_ARRAY);
 
+    if (len < 0) {
+	ArgError("negative array size (or size too big)");
+    }
+    if (len > 0 && len*sizeof(VALUE) <= 0) {
+	ArgError("array size too big");
+    }
     ary->len = 0;
     ary->capa = len;
-    if (len == 0)
-	ary->ptr = 0;
-    else {
-	ary->ptr = ALLOC_N(VALUE, len);
-	memclear(ary->ptr, len);
-    }
+    ary->ptr = 0;
+    ary->ptr = ALLOC_N(VALUE, len);
+    memclear(ary->ptr, len);
 
     return (VALUE)ary;
 }
@@ -82,30 +95,40 @@ ary_new()
     return ary_new2(ARY_DEFAULT_SIZE);
 }
 
+#ifdef HAVE_STDARG_PROTOTYPES
+#include <stdarg.h>
+#define va_init_list(a,b) va_start(a,b)
+#else
 #include <varargs.h>
+#define va_init_list(a,b) va_start(a)
+#endif
 
 VALUE
+#ifdef HAVE_STDARG_PROTOTYPES
+ary_new3(int n, ...)
+#else
 ary_new3(n, va_alist)
     int n;
     va_dcl
+#endif
 {
     va_list ar;
-    struct RArray* ary;
+    VALUE ary;
     int i;
 
     if (n < 0) {
 	IndexError("Negative number of items(%d)", n);
     }
-    ary = (struct RArray*)ary_new2(n<ARY_DEFAULT_SIZE?ARY_DEFAULT_SIZE:n);
+    ary = ary_new2(n<ARY_DEFAULT_SIZE?ARY_DEFAULT_SIZE:n);
 
-    va_start(ar);
+    va_init_list(ar, n);
     for (i=0; i<n; i++) {
-	ary->ptr[i] = va_arg(ar, VALUE);
+	RARRAY(ary)->ptr[i] = va_arg(ar, VALUE);
     }
     va_end(ar);
 
-    ary->len = n;
-    return (VALUE)ary;
+    RARRAY(ary)->len = n;
+    return ary;
 }
 
 VALUE
@@ -113,56 +136,75 @@ ary_new4(n, elts)
     int n;
     VALUE *elts;
 {
-    struct RArray* ary;
+    VALUE ary;
 
-    ary = (struct RArray*)ary_new2(n);
-    MEMCPY(ary->ptr, elts, VALUE, n);
-    ary->len = n;
+    ary = ary_new2(n);
+    if (elts) {
+	MEMCPY(RARRAY(ary)->ptr, elts, VALUE, n);
+    }
+    RARRAY(ary)->len = n;
 
-    return (VALUE)ary;
+    return ary;
 }
 
 VALUE
 assoc_new(car, cdr)
     VALUE car, cdr;
 {
-    struct RArray* ary;
+    VALUE ary;
 
-    ary = (struct RArray*)ary_new2(2);
-    ary->ptr[0] = car;
-    ary->ptr[1] = cdr;
-    ary->len = 2;
+    ary = ary_new2(2);
+    RARRAY(ary)->ptr[0] = car;
+    RARRAY(ary)->ptr[1] = cdr;
+    RARRAY(ary)->len = 2;
 
-    return (VALUE)ary;
+    return ary;
 }
 
 static VALUE
-ary_s_new(argc, argv, class)
+ary_s_new(argc, argv, klass)
     int argc;
     VALUE *argv;
-    VALUE class;
+    VALUE klass;
 {
-    VALUE size;
+    int len = 0;
+    VALUE size, val;
     NEWOBJ(ary, struct RArray);
-    OBJSETUP(ary, class, T_ARRAY);
+    OBJSETUP(ary, klass, T_ARRAY);
 
-    rb_scan_args(argc, argv, "01", &size);
     ary->len = 0;
-    ary->capa = NIL_P(size)?ARY_DEFAULT_SIZE:NUM2INT(size);
+    ary->ptr = 0;
+    if (rb_scan_args(argc, argv, "02", &size, &val) == 0) {
+	ary->capa = ARY_DEFAULT_SIZE;
+    }
+    else {
+	int capa = NUM2INT(size);
+
+	if (capa < 0) {
+	    ArgError("negative array size");
+	}
+	if (capa > 0 && capa*sizeof(VALUE) <= 0) {
+	    ArgError("array size too big");
+	}
+	ary->capa = capa;
+	len = capa;
+    }
     ary->ptr = ALLOC_N(VALUE, ary->capa);
-    memclear(ary->ptr, ary->capa);
+    memfill(ary->ptr, len, val);
+    ary->len = len;
+    obj_call_init((VALUE)ary);
 
     return (VALUE)ary;
 }
 
 static VALUE
-ary_s_create(argc, argv, class)
+ary_s_create(argc, argv, klass)
     int argc;
     VALUE *argv;
-    VALUE class;
+    VALUE klass;
 {
     NEWOBJ(ary, struct RArray);
-    OBJSETUP(ary, class, T_ARRAY);
+    OBJSETUP(ary, klass, T_ARRAY);
 
     ary->len = argc;
     ary->capa = argc;
@@ -179,7 +221,7 @@ ary_s_create(argc, argv, class)
 
 void
 ary_store(ary, idx, val)
-    struct RArray *ary;
+    VALUE ary;
     int idx;
     VALUE val;
 {
@@ -188,69 +230,69 @@ ary_store(ary, idx, val)
 	IndexError("negative index for array");
     }
 
-    if (idx >= ary->capa) {
-	ary->capa = idx + ARY_DEFAULT_SIZE;
-	REALLOC_N(ary->ptr, VALUE, ary->capa);
+    if (idx >= RARRAY(ary)->capa) {
+	RARRAY(ary)->capa = idx + ARY_DEFAULT_SIZE;
+	REALLOC_N(RARRAY(ary)->ptr, VALUE, RARRAY(ary)->capa);
     }
-    if (idx > ary->len) {
-	memclear(ary->ptr+ary->len, idx-ary->len+1);
+    if (idx > RARRAY(ary)->len) {
+	memclear(RARRAY(ary)->ptr+RARRAY(ary)->len, idx-RARRAY(ary)->len+1);
     }
 
-    if (idx >= ary->len) {
-	ary->len = idx + 1;
+    if (idx >= RARRAY(ary)->len) {
+	RARRAY(ary)->len = idx + 1;
     }
-    ary->ptr[idx] = val;
+    RARRAY(ary)->ptr[idx] = val;
 }
 
 VALUE
 ary_push(ary, item)
-    struct RArray *ary;
+    VALUE ary;
     VALUE item;
 {
-    ary_store(ary, ary->len, item);
-    return (VALUE)ary;
+    ary_store(ary, RARRAY(ary)->len, item);
+    return ary;
 }
 
 static VALUE
 ary_push_method(argc, argv, ary)
     int argc;
     VALUE *argv;
-    struct RArray *ary;
+    VALUE ary;
 {
     while (argc--) {
-	ary_store(ary, ary->len, *argv++);
+	ary_store(ary, RARRAY(ary)->len, *argv++);
     }
-    return (VALUE)ary;
+    return ary;
 }
 
 VALUE
 ary_pop(ary)
-    struct RArray *ary;
+    VALUE ary;
 {
-    if (ary->len == 0) return Qnil;
-    if (ary->len * 10 < ary->capa && ary->capa > ARY_DEFAULT_SIZE) {
-	ary->capa = ary->len * 2;
-	REALLOC_N(ary->ptr, VALUE, ary->capa);
+    if (RARRAY(ary)->len == 0) return Qnil;
+    if (RARRAY(ary)->len * 10 < RARRAY(ary)->capa && RARRAY(ary)->capa > ARY_DEFAULT_SIZE) {
+	RARRAY(ary)->capa = RARRAY(ary)->len * 2;
+	REALLOC_N(RARRAY(ary)->ptr, VALUE, RARRAY(ary)->capa);
     }
-    return ary->ptr[--ary->len];
+    return RARRAY(ary)->ptr[--RARRAY(ary)->len];
 }
 
 VALUE
 ary_shift(ary)
-    struct RArray *ary;
+    VALUE ary;
 {
     VALUE top;
 
-    if (ary->len == 0) return Qnil;
+    if (RARRAY(ary)->len == 0) return Qnil;
 
-    top = ary->ptr[0];
-    ary->len--;
+    top = RARRAY(ary)->ptr[0];
+    RARRAY(ary)->len--;
 
     /* sliding items */
-    MEMMOVE(ary->ptr, ary->ptr+1, VALUE, ary->len);
-    if (ary->len * 10 < ary->capa && ary->capa > ARY_DEFAULT_SIZE) {
-	ary->capa = ary->len * 2;
-	REALLOC_N(ary->ptr, VALUE, ary->capa);
+    MEMMOVE(RARRAY(ary)->ptr, RARRAY(ary)->ptr+1, VALUE, RARRAY(ary)->len);
+    if (RARRAY(ary)->len * 10 < RARRAY(ary)->capa && RARRAY(ary)->capa > ARY_DEFAULT_SIZE) {
+	RARRAY(ary)->capa = RARRAY(ary)->len * 2;
+	REALLOC_N(RARRAY(ary)->ptr, VALUE, RARRAY(ary)->capa);
     }
 
     return top;
@@ -258,65 +300,67 @@ ary_shift(ary)
 
 VALUE
 ary_unshift(ary, item)
-    struct RArray *ary;
-    int item;
+    VALUE ary, item;
 {
     ary_modify(ary);
-    if (ary->len >= ary->capa) {
-	ary->capa+=ARY_DEFAULT_SIZE;
-	REALLOC_N(ary->ptr, VALUE, ary->capa);
+    if (RARRAY(ary)->len >= RARRAY(ary)->capa) {
+	RARRAY(ary)->capa+=ARY_DEFAULT_SIZE;
+	REALLOC_N(RARRAY(ary)->ptr, VALUE, RARRAY(ary)->capa);
     }
 
     /* sliding items */
-    MEMMOVE(ary->ptr+1, ary->ptr, VALUE, ary->len);
+    MEMMOVE(RARRAY(ary)->ptr+1, RARRAY(ary)->ptr, VALUE, RARRAY(ary)->len);
 
-    ary->len++;
-    return ary->ptr[0] = item;
+    RARRAY(ary)->len++;
+    return RARRAY(ary)->ptr[0] = item;
 }
 
 VALUE
 ary_entry(ary, offset)
-    struct RArray *ary;
+    VALUE ary;
     int offset;
 {
-    if (ary->len == 0) return Qnil;
+    if (RARRAY(ary)->len == 0) return Qnil;
 
     if (offset < 0) {
-	offset = ary->len + offset;
+	offset = RARRAY(ary)->len + offset;
     }
-    if (offset < 0 || ary->len <= offset) {
+    if (offset < 0 || RARRAY(ary)->len <= offset) {
 	return Qnil;
     }
 
-    return ary->ptr[offset];
+    return RARRAY(ary)->ptr[offset];
 }
 
 static VALUE
 ary_subseq(ary, beg, len)
-    struct RArray *ary;
+    VALUE ary;
     int beg, len;
 {
-    struct RArray *ary2;
+    VALUE ary2;
 
     if (beg < 0) {
-	beg = ary->len + beg;
+	beg = RARRAY(ary)->len + beg;
 	if (beg < 0) beg = 0;
     }
     if (len < 0) {
-	IndexError("negative length %d", ary->len);
+	IndexError("negative length %d", RARRAY(ary)->len);
     }
     if (len == 0) {
 	return ary_new2(0);
     }
-    if (beg + len > ary->len) {
-	len = ary->len - beg;
+    if (beg + len > RARRAY(ary)->len) {
+	len = RARRAY(ary)->len - beg;
+    }
+    if (len < 0) {
+	len = 0;
     }
 
-    ary2 = (struct RArray*)ary_new2(len);
-    MEMCPY(ary2->ptr, ary->ptr+beg, VALUE, len);
-    ary2->len = len;
+    ary2 = ary_new2(len);
+    MEMCPY(RARRAY(ary2)->ptr, RARRAY(ary)->ptr+beg, VALUE, len);
+    RARRAY(ary2)->len = len;
 
-    return (VALUE)ary2;
+    return ary2;
 }
 
 static VALUE
@@ -326,42 +370,38 @@ beg_len(range, begp, lenp, len)
     int len;
 {
     int beg, end;
+    int b, e;
 
     if (!range_beg_end(range, &beg, &end)) return FALSE;
-
-    if ((beg > 0 && end > 0 || beg < 0 && end < 0) && beg > end) {
-	IndexError("end smaller than beg [%d..%d]", beg, end);
-    }
+    b = beg; e = end;
 
     if (beg < 0) {
 	beg = len + beg;
 	if (beg < 0) beg = 0;
     }
+    if (end < 0) {
+	end = len + end;
+	if (end < 0) end = -1;
+    }
+    if (beg > end) {
+	IndexError("end smaller than beg [%d..%d]", b, e);
+    }
+
     *begp = beg;
     if (beg > len) {
 	*lenp = 0;
     }
     else {
-	if (end < 0) {
-	    end = len + end;
-	    if (end < 0) end = -1;
-	}
-	if (end > len) end = len;
-	if (beg > end) {
-	    *lenp = 0;
-	}
-	else {
-	    *lenp = end - beg +1;
-	}
+	*lenp = end - beg +1;
     }
     return TRUE;
 }
 
-static VALUE
+VALUE
 ary_aref(argc, argv, ary)
     int argc;
     VALUE *argv;
-    struct RArray *ary;
+    VALUE ary;
 {
     VALUE arg1, arg2;
     int beg, len;
@@ -379,99 +419,106 @@ ary_aref(argc, argv, ary)
     if (FIXNUM_P(arg1)) {
 	return ary_entry(ary, FIX2INT(arg1));
     }
-    else {
-	/* check if idx is Range */
-	if (beg_len(arg1, &beg, &len, ary->len)) {
-	    return ary_subseq(ary, beg, len);
-	}
-    }
-    if (TYPE(arg1) == T_BIGNUM) {
+    else if (TYPE(arg1) == T_BIGNUM) {
 	IndexError("index too big");
     }
+    else if (beg_len(arg1, &beg, &len, RARRAY(ary)->len)) {
+	/* check if idx is Range */
+	return ary_subseq(ary, beg, len);
+    }
     return ary_entry(ary, NUM2INT(arg1));
 }
 
 static VALUE
 ary_index(ary, val)
-    struct RArray *ary;
+    VALUE ary;
     VALUE val;
 {
     int i;
 
-    for (i=0; i<ary->len; i++) {
-	if (rb_equal(ary->ptr[i], val))
+    for (i=0; i<RARRAY(ary)->len; i++) {
+	if (rb_equal(RARRAY(ary)->ptr[i], val))
 	    return INT2FIX(i);
     }
     return Qnil;
 }
 
 static VALUE
-ary_indexes(ary, args)
-    struct RArray *ary, *args;
+ary_rindex(ary, val)
+    VALUE ary;
+    VALUE val;
+{
+    int i = RARRAY(ary)->len;
+
+    while (i--) {
+	if (rb_equal(RARRAY(ary)->ptr[i], val))
+	    return INT2FIX(i);
+    }
+    return Qnil;
+}
+
+static VALUE
+ary_indexes(argc, argv, ary)
+    int argc;
+    VALUE *argv;
+    VALUE ary;
 {
-    VALUE *p, *pend;
     VALUE new_ary;
-    int i = 0;
+    int i;
 
-    if (!args || NIL_P(args)) {
-	return ary_new2(0);
+    new_ary = ary_new2(argc);
+    for (i=0; i<argc; i++) {
+	ary_store(new_ary, i, ary_entry(ary, NUM2INT(argv[i])));
     }
 
-    new_ary = ary_new2(args->len);
-
-    p = args->ptr; pend = p + args->len;
-    while (p < pend) {
-	ary_store(new_ary, i++, ary_entry(ary, NUM2INT(*p)));
-	p++;
-    }
     return new_ary;
 }
 
 static void
 ary_replace(ary, beg, len, rpl)
-    struct RArray *ary, *rpl;
+    VALUE ary, rpl;
     int beg, len;
 {
     ary_modify(ary);
     if (TYPE(rpl) != T_ARRAY) {
-	rpl = (struct RArray*)rb_to_a(rpl);
+	rpl = rb_Array(rpl);
     }
     if (beg < 0) {
-	beg = ary->len + beg;
+	beg = RARRAY(ary)->len + beg;
 	if (beg < 0) beg = 0;
     }
-    if (beg >= ary->len) {
-	len = beg + rpl->len;
-	if (len >= ary->capa) {
-	    ary->capa=len;
-	    REALLOC_N(ary->ptr, VALUE, ary->capa);
+    if (beg >= RARRAY(ary)->len) {
+	len = beg + RARRAY(rpl)->len;
+	if (len >= RARRAY(ary)->capa) {
+	    RARRAY(ary)->capa=len;
+	    REALLOC_N(RARRAY(ary)->ptr, VALUE, RARRAY(ary)->capa);
 	}
-	memclear(ary->ptr+ary->len, beg-ary->len);
-	MEMCPY(ary->ptr+beg, rpl->ptr, VALUE, rpl->len);
-	ary->len = len;
+	memclear(RARRAY(ary)->ptr+RARRAY(ary)->len, beg-RARRAY(ary)->len);
+	MEMCPY(RARRAY(ary)->ptr+beg, RARRAY(rpl)->ptr, VALUE, RARRAY(rpl)->len);
+	RARRAY(ary)->len = len;
     }
     else {
 	int alen;
 
-	if (beg + len > ary->len) {
-	    len = ary->len - beg;
+	if (beg + len > RARRAY(ary)->len) {
+	    len = RARRAY(ary)->len - beg;
 	}
 	if (len < 0) {
-	    IndexError("negative length %d", ary->len);
+	    IndexError("negative length %d", RARRAY(ary)->len);
 	}
 
-	alen = ary->len + rpl->len - len;
-	if (alen >= ary->capa) {
-	    ary->capa=alen;
-	    REALLOC_N(ary->ptr, VALUE, ary->capa);
+	alen = RARRAY(ary)->len + RARRAY(rpl)->len - len;
+	if (alen >= RARRAY(ary)->capa) {
+	    RARRAY(ary)->capa=alen;
+	    REALLOC_N(RARRAY(ary)->ptr, VALUE, RARRAY(ary)->capa);
 	}
 
 	if (len != RARRAY(rpl)->len) {
-	    MEMMOVE(ary->ptr+beg+rpl->len, ary->ptr+beg+len,
-		    VALUE, ary->len-(beg+len));
-	    ary->len = alen;
+	    MEMMOVE(RARRAY(ary)->ptr+beg+RARRAY(rpl)->len, RARRAY(ary)->ptr+beg+len,
+		    VALUE, RARRAY(ary)->len-(beg+len));
+	    RARRAY(ary)->len = alen;
 	}
-	MEMCPY(ary->ptr+beg, rpl->ptr, VALUE, rpl->len);
+	MEMCPY(RARRAY(ary)->ptr+beg, RARRAY(rpl)->ptr, VALUE, RARRAY(rpl)->len);
     }
 }
 
@@ -479,10 +526,9 @@ static VALUE
 ary_aset(argc, argv, ary)
     int argc;
     VALUE *argv;
-    struct RArray *ary;
+    VALUE ary;
 {
-    VALUE arg1, arg2;
-    struct RArray *arg3;
+    VALUE arg1, arg2, arg3;
     int offset;
     int beg, len;
 
@@ -490,13 +536,13 @@ ary_aset(argc, argv, ary)
 	beg = NUM2INT(arg1);
 	len = NUM2INT(arg2);
 	ary_replace(ary, beg, len, arg3);
-	return (VALUE)arg3;
+	return arg3;
     }
     else if (FIXNUM_P(arg1)) {
 	offset = FIX2INT(arg1);
 	goto fixnum;
     }
-    else if (beg_len(arg1, &beg, &len, ary->len)) {
+    else if (beg_len(arg1, &beg, &len, RARRAY(ary)->len)) {
 	/* check if idx is Range */
 	ary_replace(ary, beg, len, arg2);
 	return arg2;
@@ -508,7 +554,7 @@ ary_aset(argc, argv, ary)
     offset = NUM2INT(arg1);
   fixnum:
     if (offset < 0) {
-	offset = ary->len + offset;
+	offset = RARRAY(ary)->len + offset;
     }
     ary_store(ary, offset, arg2);
     return arg2;
@@ -516,23 +562,23 @@ ary_aset(argc, argv, ary)
 
 VALUE
 ary_each(ary)
-    struct RArray *ary;
+    VALUE ary;
 {
     int i;
 
-    for (i=0; i<ary->len; i++) {
-	rb_yield(ary->ptr[i]);
+    for (i=0; i<RARRAY(ary)->len; i++) {
+	rb_yield(RARRAY(ary)->ptr[i]);
     }
     return Qnil;
 }
 
 static VALUE
 ary_each_index(ary)
-    struct RArray *ary;
+    VALUE ary;
 {
     int i;
 
-    for (i=0; i<ary->len; i++) {
+    for (i=0; i<RARRAY(ary)->len; i++) {
 	rb_yield(INT2FIX(i));
     }
     return Qnil;
@@ -540,69 +586,83 @@ ary_each_index(ary)
 
 static VALUE
 ary_reverse_each(ary)
-    struct RArray *ary;
+    VALUE ary;
 {
-    int len = ary->len;
+    int len = RARRAY(ary)->len;
 
     while (len--) {
-	rb_yield(ary->ptr[len]);
+	rb_yield(RARRAY(ary)->ptr[len]);
     }
     return Qnil;
 }
 
 static VALUE
 ary_length(ary)
-    struct RArray *ary;
+    VALUE ary;
 {
-    return INT2FIX(ary->len);
+    return INT2FIX(RARRAY(ary)->len);
 }
 
 static VALUE
 ary_empty_p(ary)
-    struct RArray *ary;
+    VALUE ary;
 {
-    if (ary->len == 0)
+    if (RARRAY(ary)->len == 0)
 	return TRUE;
     return FALSE;
 }
 
 static VALUE
 ary_clone(ary)
-    struct RArray *ary;
+    VALUE ary;
 {
-    VALUE ary2 = ary_new2(ary->len);
+    VALUE ary2 = ary_new2(RARRAY(ary)->len);
 
     CLONESETUP(ary2, ary);
-    MEMCPY(RARRAY(ary2)->ptr, ary->ptr, VALUE, ary->len);
-    RARRAY(ary2)->len = ary->len;
+    MEMCPY(RARRAY(ary2)->ptr, RARRAY(ary)->ptr, VALUE, RARRAY(ary)->len);
+    RARRAY(ary2)->len = RARRAY(ary)->len;
     return ary2;
 }
 
+static VALUE
+ary_dup(ary)
+    VALUE ary;
+{
+    return ary_new4(RARRAY(ary)->len, RARRAY(ary)->ptr);
+}
+
+static VALUE
+to_ary(ary)
+    VALUE ary;
+{
+    return rb_convert_type(ary, T_ARRAY, "Array", "to_ary");
+}
+
 extern VALUE OFS;
 
 VALUE
 ary_join(ary, sep)
-    struct RArray *ary;
-    struct RString *sep;
+    VALUE ary;
+    VALUE sep;
 {
     int i;
     VALUE result, tmp;
-    if (ary->len == 0) return str_new(0, 0);
+    if (RARRAY(ary)->len == 0) return str_new(0, 0);
 
-    switch (TYPE(ary->ptr[0])) {
+    switch (TYPE(RARRAY(ary)->ptr[0])) {
       case T_STRING:
-	result = str_dup(ary->ptr[0]);
+	result = str_dup(RARRAY(ary)->ptr[0]);
 	break;
       case T_ARRAY:
-	result = ary_join(ary->ptr[0], sep);
+	result = ary_join(RARRAY(ary)->ptr[0], sep);
 	break;
       default:
-	result = obj_as_string(ary->ptr[0]);
+	result = obj_as_string(RARRAY(ary)->ptr[0]);
 	break;
     }
 
-    for (i=1; i<ary->len; i++) {
-	tmp = ary->ptr[i];
+    for (i=1; i<RARRAY(ary)->len; i++) {
+	tmp = RARRAY(ary)->ptr[i];
 	switch (TYPE(tmp)) {
 	  case T_STRING:
 	    break;
@@ -612,7 +672,7 @@ ary_join(ary, sep)
 	  default:
 	    tmp = obj_as_string(tmp);
 	}
-	if (!NIL_P(sep)) str_cat(result, sep->ptr, sep->len);
+	if (!NIL_P(sep)) str_concat(result, sep);
 	str_cat(result, RSTRING(tmp)->ptr, RSTRING(tmp)->len);
 	if (str_tainted(tmp)) str_taint(result);
     }
@@ -624,13 +684,12 @@ static VALUE
 ary_join_method(argc, argv, ary)
     int argc;
     VALUE *argv;
-    struct RArray *ary;
+    VALUE ary;
 {
     VALUE sep;
 
     rb_scan_args(argc, argv, "01", &sep);
     if (NIL_P(sep)) sep = OFS;
-    if (!NIL_P(sep)) Check_Type(sep, T_STRING);
 
     return ary_join(ary, sep);
 }
@@ -644,35 +703,19 @@ ary_to_s(ary)
     return str;
 }
 
-VALUE
-ary_print_on(ary, port)
-    struct RArray *ary;
-    VALUE port;
-{
-    int i;
-
-    for (i=0; i<ary->len; i++) {
-	if (!NIL_P(OFS) && i>0) {
-	    io_write(port, OFS);
-	}
-	io_write(port, ary->ptr[i]);
-    }
-    return port;
-}
-
 static VALUE
 ary_inspect(ary)
-    struct RArray *ary;
+    VALUE ary;
 {
     int i, len;
     VALUE s, str;
 
-    if (ary->len == 0) return str_new2("[]");
+    if (RARRAY(ary)->len == 0) return str_new2("[]");
     str = str_new2("[");
     len = 1;
 
-    for (i=0; i<ary->len; i++) {
-	s = rb_inspect(ary->ptr[i]);
+    for (i=0; i<RARRAY(ary)->len; i++) {
+	s = rb_inspect(RARRAY(ary)->ptr[i]);
 	if (i > 0) str_cat(str, ", ", 2);
 	str_cat(str, RSTRING(s)->ptr, RSTRING(s)->len);
 	len += RSTRING(s)->len + 2;
@@ -689,27 +732,17 @@ ary_to_a(ary)
     return ary;
 }
 
-VALUE
-rb_to_a(obj)
-    VALUE obj;
-{
-    if (TYPE(obj) == T_ARRAY) return obj;
-    obj = rb_funcall(obj, rb_intern("to_a"), 0);
-    if (TYPE(obj) != T_ARRAY) {
-	Bug("`to_a' did not return Array");
-    }
-    return obj;
-}
-
 VALUE
 ary_reverse(ary)
-    struct RArray *ary;
+    VALUE ary;
 {
     VALUE *p1, *p2;
     VALUE tmp;
 
-    p1 = ary->ptr;
-    p2 = p1 + ary->len - 1;	/* points last item */
+    if (RARRAY(ary)->len == 0) return ary;
+
+    p1 = RARRAY(ary)->ptr;
+    p2 = p1 + RARRAY(ary)->len - 1;	/* points last item */
 
     while (p1 < p2) {
 	tmp = *p1;
@@ -718,14 +751,14 @@ ary_reverse(ary)
 	p1++; p2--;
     }
 
-    return (VALUE)ary;
+    return ary;
 }
 
 static VALUE
 ary_reverse_method(ary)
-    struct RArray *ary;
+    VALUE ary;
 {
-    return ary_reverse(ary_clone(ary));
+    return ary_reverse(ary_dup(ary));
 }
 
 static ID cmp;
@@ -757,41 +790,47 @@ sort_2(a, b)
 
 VALUE
 ary_sort_bang(ary)
-    struct RArray *ary;
+    VALUE ary;
 {
+    if (RARRAY(ary)->len == 0) return ary;
+
     ary_modify(ary);
-    qsort(ary->ptr, ary->len, sizeof(VALUE), iterator_p()?sort_1:sort_2);
-    return (VALUE)ary;
+    rb_qsort(RARRAY(ary)->ptr, RARRAY(ary)->len, sizeof(VALUE),
+	  iterator_p()?sort_1:sort_2);
+    return ary;
 }
 
 VALUE
 ary_sort(ary)
     VALUE ary;
 {
-    return ary_sort_bang(ary_clone(ary));
+    if (RARRAY(ary)->len == 0) return ary;
+    return ary_sort_bang(ary_dup(ary));
 }
 
 VALUE
 ary_delete(ary, item)
-    struct RArray *ary;
+    VALUE ary;
     VALUE item;
 {
     int i1, i2;
 
     ary_modify(ary);
-    for (i1 = i2 = 0; i1 < ary->len; i1++) {
-	if (rb_equal(ary->ptr[i1], item)) continue;
+    for (i1 = i2 = 0; i1 < RARRAY(ary)->len; i1++) {
+	if (rb_equal(RARRAY(ary)->ptr[i1], item)) continue;
 	if (i1 != i2) {
-	    ary->ptr[i2] = ary->ptr[i1];
+	    RARRAY(ary)->ptr[i2] = RARRAY(ary)->ptr[i1];
 	}
 	i2++;
     }
-    if (ary->len == i2) {
-	if (iterator_p()) rb_yield(item);
+    if (RARRAY(ary)->len == i2) {
+	if (iterator_p()) {
+	    return rb_yield(item);
+	}
 	return Qnil;
     }
     else {
-	ary->len = i2;
+	RARRAY(ary)->len = i2;
     }
 
     return item;
@@ -799,7 +838,7 @@ ary_delete(ary, item)
 
 VALUE
 ary_delete_at(ary, at)
-    struct RArray *ary;
+    VALUE ary;
     VALUE at;
 {
     int i1, i2, pos;
@@ -807,184 +846,187 @@ ary_delete_at(ary, at)
 
     ary_modify(ary);
     pos = NUM2INT(at);
-    for (i1 = i2 = 0; i1 < ary->len; i1++) {
+    for (i1 = i2 = 0; i1 < RARRAY(ary)->len; i1++) {
 	if (i1 == pos) {
-	    del = ary->ptr[i1];
+	    del = RARRAY(ary)->ptr[i1];
 	    continue;
 	}
 	if (i1 != i2) {
-	    ary->ptr[i2] = ary->ptr[i1];
+	    RARRAY(ary)->ptr[i2] = RARRAY(ary)->ptr[i1];
 	}
 	i2++;
     }
-    ary->len = i2;
+    RARRAY(ary)->len = i2;
 
     return del;
 }
 
 static VALUE
 ary_delete_if(ary)
-    struct RArray *ary;
+    VALUE ary;
 {
     int i1, i2;
 
     ary_modify(ary);
-    for (i1 = i2 = 0; i1 < ary->len; i1++) {
-	if (rb_yield(ary->ptr[i1])) continue;
+    for (i1 = i2 = 0; i1 < RARRAY(ary)->len; i1++) {
+	if (rb_yield(RARRAY(ary)->ptr[i1])) continue;
 	if (i1 != i2) {
-	    ary->ptr[i2] = ary->ptr[i1];
+	    RARRAY(ary)->ptr[i2] = RARRAY(ary)->ptr[i1];
 	}
 	i2++;
     }
-    ary->len = i2;
+    RARRAY(ary)->len = i2;
 
-    return (VALUE)ary;
+    return ary;
 }
 
-#if 0
 static VALUE
-ary_replace(ary)
-    struct RArray *ary;
+ary_filter(ary)
+    VALUE ary;
 {
     int i;
 
-    for (i = 0; i < ary->len; i++) {
-	ary->ptr[i] = rb_yield(ary->ptr[i]);
+    ary_modify(ary);
+    for (i = 0; i < RARRAY(ary)->len; i++) {
+	RARRAY(ary)->ptr[i] = rb_yield(RARRAY(ary)->ptr[i]);
     }
+    return ary;
+}
 
-    return (VALUE)ary;
+static VALUE
+ary_replace_method(ary, ary2)
+    VALUE ary, ary2;
+{
+    ary2 = to_ary(ary2);
+    ary_replace(ary, 0, RARRAY(ary2)->len, ary2);
+    return ary;
 }
-#endif
 
 static VALUE
 ary_clear(ary)
-    struct RArray *ary;
+    VALUE ary;
 {
-    ary->len = 0;
-    if (ARY_DEFAULT_SIZE*3 < ary->capa) {
-	ary->capa = ARY_DEFAULT_SIZE * 2;
-	REALLOC_N(ary->ptr, VALUE, ary->capa);
+    RARRAY(ary)->len = 0;
+    if (ARY_DEFAULT_SIZE*3 < RARRAY(ary)->capa) {
+	RARRAY(ary)->capa = ARY_DEFAULT_SIZE * 2;
+	REALLOC_N(RARRAY(ary)->ptr, VALUE, RARRAY(ary)->capa);
     }
-    return (VALUE)ary;
+    return ary;
 }
 
 static VALUE
 ary_fill(argc, argv, ary)
     int argc;
     VALUE *argv;
-    struct RArray *ary;
+    VALUE ary;
 {
     VALUE item, arg1, arg2;
     int beg, len, end;
     VALUE *p, *pend;
 
-    rb_scan_args(argc, argv, "12", &item, &arg1, &arg2);
-    if (NIL_P(arg2) && beg_len(arg1, &beg, &len, ary->len)) {
+    if (rb_scan_args(argc, argv, "12", &item, &arg1, &arg2) == 2 &&
+	beg_len(arg1, &beg, &len, RARRAY(ary)->len)) {
 	/* beg and len set already */
     }
     else {
-	beg = NUM2INT(arg1);
+	beg = NIL_P(arg1)?0:NUM2INT(arg1);
 	if (beg < 0) {
-	    beg = ary->len + beg;
+	    beg = RARRAY(ary)->len + beg;
 	    if (beg < 0) beg = 0;
 	}
-	if (arg2) {
-	    len = NUM2INT(arg2);
-	}
-	else {
-	    len = ary->len - beg;
-	}
+	len = NIL_P(arg2)?RARRAY(ary)->len - beg:NUM2INT(arg2);
     }
     end = beg + len;
-    if (end > ary->len) {
-	if (end >= ary->capa) {
-	    ary->capa=end;
-	    REALLOC_N(ary->ptr, VALUE, ary->capa);
+    if (end > RARRAY(ary)->len) {
+	if (end >= RARRAY(ary)->capa) {
+	    RARRAY(ary)->capa=end;
+	    REALLOC_N(RARRAY(ary)->ptr, VALUE, RARRAY(ary)->capa);
 	}
-	if (beg > ary->len) {
-	    memclear(ary->ptr+ary->len, end-ary->len);
+	if (beg > RARRAY(ary)->len) {
+	    memclear(RARRAY(ary)->ptr+RARRAY(ary)->len,end-RARRAY(ary)->len);
 	}
-	ary->len = end;
+	RARRAY(ary)->len = end;
     }
-    p = ary->ptr + beg; pend = p + len;
+    p = RARRAY(ary)->ptr + beg; pend = p + len;
 
     while (p < pend) {
 	*p++ = item;
     }
-    return (VALUE)ary;
+    return ary;
 }
 
 VALUE
 ary_plus(x, y)
-    struct RArray *x, *y;
+    VALUE x, y;
 {
-    struct RArray *z;
+    VALUE z;
 
     if (TYPE(y) != T_ARRAY) {
-	return ary_plus(x, rb_to_a(y));
+	return ary_plus(x, rb_Array(y));
     }
 
-    z = (struct RArray*)ary_new2(x->len + y->len);
-    MEMCPY(z->ptr, x->ptr, VALUE, x->len);
-    MEMCPY(z->ptr+x->len, y->ptr, VALUE, y->len);
-    z->len = x->len + RARRAY(y)->len;
-    return (VALUE)z;
+    z = ary_new2(RARRAY(x)->len + RARRAY(y)->len);
+    MEMCPY(RARRAY(z)->ptr, RARRAY(x)->ptr, VALUE, RARRAY(x)->len);
+    MEMCPY(RARRAY(z)->ptr+RARRAY(x)->len, RARRAY(y)->ptr, VALUE, RARRAY(y)->len);
+    RARRAY(z)->len = RARRAY(x)->len + RARRAY(y)->len;
+    return z;
 }
 
 VALUE
 ary_concat(x, y)
-    struct RArray *x, *y;
+    VALUE x, y;
 {
     VALUE *p, *pend;
 
     if (TYPE(y) != T_ARRAY) {
-	return ary_concat(x, rb_to_a(y));
+	return ary_concat(x, rb_Array(y));
     }
 
-    p = y->ptr;
-    pend = p + y->len;
+    p = RARRAY(y)->ptr;
+    pend = p + RARRAY(y)->len;
     while (p < pend) {
-	ary_store(x, x->len, *p);
+	ary_store(x, RARRAY(x)->len, *p);
 	p++;
     }
-    return (VALUE)x;
+    return x;
 }
 
 static VALUE
 ary_times(ary, times)
-    struct RArray *ary;
+    VALUE ary;
     VALUE times;
 {
-    struct RArray *ary2;
+    VALUE ary2;
     int i, len;
 
     if (TYPE(times) == T_STRING) {
 	return ary_join(ary, times);
     }
 
-    len = NUM2INT(times) * ary->len;
-    ary2 = (struct RArray*)ary_new2(len);
-    ary2->len = len;
-
+    len = NUM2INT(times);
     if (len < 0) {
 	ArgError("negative argument");
     }
+    len *= RARRAY(ary)->len;
 
-    for (i=0; i<len; i+=ary->len) {
-	MEMCPY(ary2->ptr+i, ary->ptr, VALUE, ary->len);
+    ary2 = ary_new2(len);
+    RARRAY(ary2)->len = len;
+
+    for (i=0; i<len; i+=RARRAY(ary)->len) {
+	MEMCPY(RARRAY(ary2)->ptr+i, RARRAY(ary)->ptr, VALUE, RARRAY(ary)->len);
     }
 
-    return (VALUE)ary2;
+    return ary2;
 }
 
 VALUE
 ary_assoc(ary, key)
-    struct RArray *ary;
+    VALUE ary;
     VALUE key;
 {
     VALUE *p, *pend;
 
-    p = ary->ptr; pend = p + ary->len;
+    p = RARRAY(ary)->ptr; pend = p + RARRAY(ary)->len;
     while (p < pend) {
 	if (TYPE(*p) == T_ARRAY
 	    && RARRAY(*p)->len > 1
@@ -997,12 +1039,12 @@ ary_assoc(ary, key)
 
 VALUE
 ary_rassoc(ary, value)
-    struct RArray *ary;
+    VALUE ary;
     VALUE value;
 {
     VALUE *p, *pend;
 
-    p = ary->ptr; pend = p + ary->len;
+    p = RARRAY(ary)->ptr; pend = p + RARRAY(ary)->len;
     while (p < pend) {
 	if (TYPE(*p) == T_ARRAY
 	    && RARRAY(*p)->len > 1
@@ -1015,14 +1057,14 @@ ary_rassoc(ary, value)
 
 static VALUE
 ary_equal(ary1, ary2)
-    struct RArray *ary1, *ary2;
+    VALUE ary1, ary2;
 {
     int i;
 
     if (TYPE(ary2) != T_ARRAY) return FALSE;
-    if (ary1->len != ary2->len) return FALSE;
-    for (i=0; i<ary1->len; i++) {
-	if (!rb_equal(ary1->ptr[i], ary2->ptr[i]))
+    if (RARRAY(ary1)->len != RARRAY(ary2)->len) return FALSE;
+    for (i=0; i<RARRAY(ary1)->len; i++) {
+	if (!rb_equal(RARRAY(ary1)->ptr[i], RARRAY(ary2)->ptr[i]))
 	    return FALSE;
     }
     return TRUE;
@@ -1030,14 +1072,15 @@ ary_equal(ary1, ary2)
 
 static VALUE
 ary_eql(ary1, ary2)
-    struct RArray *ary1, *ary2;
+    VALUE ary1, ary2;
 {
     int i;
 
     if (TYPE(ary2) != T_ARRAY) return FALSE;
-    if (ary1->len != ary2->len) return FALSE;
-    for (i=0; i<ary1->len; i++) {
-	if (!rb_eql(ary1->ptr[i], ary2->ptr[i]))
+    if (RARRAY(ary1)->len != RARRAY(ary2)->len)
+	return FALSE;
+    for (i=0; i<RARRAY(ary1)->len; i++) {
+	if (!rb_eql(RARRAY(ary1)->ptr[i], RARRAY(ary2)->ptr[i]))
 	    return FALSE;
     }
     return TRUE;
@@ -1045,61 +1088,85 @@ ary_eql(ary1, ary2)
 
 static VALUE
 ary_hash(ary)
-    struct RArray *ary;
+    VALUE ary;
 {
     int h, i;
 
-    h = ary->len;
-    for (i=0; i<ary->len; i++) {
-	h ^= rb_hash(ary->ptr[i]);
+    h = RARRAY(ary)->len;
+    for (i=0; i<RARRAY(ary)->len; i++) {
+	h ^= rb_hash(RARRAY(ary)->ptr[i]);
     }
     return INT2FIX(h);
 }
 
 VALUE
 ary_includes(ary, item)
-    struct RArray *ary;
+    VALUE ary;
     VALUE item;
 {
     int i;
-    for (i=0; i<ary->len; i++) {
-	if (rb_equal(ary->ptr[i], item)) {
+    for (i=0; i<RARRAY(ary)->len; i++) {
+	if (rb_equal(RARRAY(ary)->ptr[i], item)) {
 	    return TRUE;
 	}
     }
     return FALSE;
 }
 
+VALUE
+ary_cmp(ary, ary2)
+    VALUE ary;
+    VALUE ary2;
+{
+    int i, len;
+
+    ary2 = to_ary(ary2);
+    len = RARRAY(ary)->len;
+    if (len > RARRAY(ary2)->len) {
+	len = RARRAY(ary2)->len;
+    }
+    for (i=0; i<len; i++) {
+	VALUE v = rb_funcall(RARRAY(ary)->ptr[i],cmp,1,RARRAY(ary2)->ptr[i]);
+	if (v != INT2FIX(0)) {
+	    return v;
+	}
+    }
+    len = RARRAY(ary)->len - RARRAY(ary2)->len;
+    if (len == 0) return INT2FIX(0);
+    if (len > 0) return INT2FIX(1);
+    return INT2FIX(-1);
+}
+
 static VALUE
 ary_diff(ary1, ary2)
-    struct RArray *ary1, *ary2;
+    VALUE ary1, ary2;
 {
     VALUE ary3;
     int i;
 
-    Check_Type(ary2, T_ARRAY);
+    ary2 = to_ary(ary2);
     ary3 = ary_new();
-    for (i=0; i<ary1->len; i++) {
-	if (ary_includes(ary2, ary1->ptr[i])) continue;
-	if (ary_includes(ary3, ary1->ptr[i])) continue;
-	ary_push(ary3, ary1->ptr[i]);
+    for (i=0; i<RARRAY(ary1)->len; i++) {
+	if (ary_includes(ary2, RARRAY(ary1)->ptr[i])) continue;
+	if (ary_includes(ary3, RARRAY(ary1)->ptr[i])) continue;
+	ary_push(ary3, RARRAY(ary1)->ptr[i]);
     }
     return ary3;
 }
 
 static VALUE
 ary_and(ary1, ary2)
-    struct RArray *ary1, *ary2;
+    VALUE ary1, ary2;
 {
     VALUE ary3;
     int i;
 
-    Check_Type(ary2, T_ARRAY);
+    ary2 = to_ary(ary2);
     ary3 = ary_new();
-    for (i=0; i<ary1->len; i++) {
-	if (ary_includes(ary2, ary1->ptr[i])
-	    && !ary_includes(ary3, ary1->ptr[i])) {
-	    ary_push(ary3, ary1->ptr[i]);
+    for (i=0; i<RARRAY(ary1)->len; i++) {
+	if (ary_includes(ary2, RARRAY(ary1)->ptr[i])
+	    && !ary_includes(ary3, RARRAY(ary1)->ptr[i])) {
+	    ary_push(ary3, RARRAY(ary1)->ptr[i]);
 	}
     }
     return ary3;
@@ -1107,63 +1174,108 @@ ary_and(ary1, ary2)
 
 static VALUE
 ary_or(ary1, ary2)
-    struct RArray *ary1, *ary2;
+    VALUE ary1, ary2;
 {
     VALUE ary3;
     int i;
 
     if (TYPE(ary2) != T_ARRAY) {
-	if (ary_includes(ary1, ary2)) return (VALUE)ary1;
+	if (ary_includes(ary1, ary2)) return ary1;
 	else return ary_plus(ary1, ary2);
     }
 
     ary3 = ary_new();
-    for (i=0; i<ary1->len; i++) {
-	if (!ary_includes(ary3, ary1->ptr[i]))
-		ary_push(ary3, ary1->ptr[i]);
+    for (i=0; i<RARRAY(ary1)->len; i++) {
+	if (!ary_includes(ary3, RARRAY(ary1)->ptr[i]))
+		ary_push(ary3, RARRAY(ary1)->ptr[i]);
     }
-    for (i=0; i<ary2->len; i++) {
-	if (!ary_includes(ary3, ary2->ptr[i]))
-		ary_push(ary3, ary2->ptr[i]);
+    for (i=0; i<RARRAY(ary2)->len; i++) {
+	if (!ary_includes(ary3, RARRAY(ary2)->ptr[i]))
+		ary_push(ary3, RARRAY(ary2)->ptr[i]);
     }
     return ary3;
 }
 
+static VALUE
+ary_uniq_bang(ary)
+    VALUE ary;
+{
+    VALUE *p, *q, *t, *end;
+    VALUE v;
+
+    ary_modify(ary);
+    p = RARRAY(ary)->ptr;
+    end = p + RARRAY(ary)->len;
+
+    while (p < end) {
+	v = *p++;
+	q = t = p;
+	while (q < end) {
+	    if (rb_equal(*q, v)) q++;
+	    else *t++ = *q++;
+	}
+	end = t;
+    }
+    if (RARRAY(ary)->len == (end - RARRAY(ary)->ptr)) {
+	return Qnil;
+    }
+
+    RARRAY(ary)->len = (end - RARRAY(ary)->ptr);
+
+    return ary;
+}
+
+static VALUE
+ary_uniq(ary)
+    VALUE ary;
+{
+    VALUE v = ary_uniq_bang(ary_dup(ary));
+
+    if (NIL_P(v)) return ary;
+    return v;
+}
+
 static VALUE
 ary_compact_bang(ary)
-    struct RArray *ary;
+    VALUE ary;
 {
     VALUE *p, *t, *end;
 
     ary_modify(ary);
-    p = t = ary->ptr;
-    end = p + ary->len;
+    p = t = RARRAY(ary)->ptr;
+    end = p + RARRAY(ary)->len;
     while (t < end) {
 	if (NIL_P(*t)) t++;
 	else *p++ = *t++;
     }
-    ary->len = ary->capa = (p - ary->ptr);
-    REALLOC_N(ary->ptr, VALUE, ary->len);
+    if (RARRAY(ary)->len == (p - RARRAY(ary)->ptr)) {
+	return Qnil;
+    }
+    RARRAY(ary)->len = RARRAY(ary)->capa = (p - RARRAY(ary)->ptr);
+    REALLOC_N(RARRAY(ary)->ptr, VALUE, RARRAY(ary)->len);
 
-    return (VALUE)ary;
+    return ary;
 }
 
 static VALUE
 ary_compact(ary)
-    struct RArray *ary;
+    VALUE ary;
 {
-    return ary_compact_bang(ary_clone(ary));
+    VALUE v = ary_compact_bang(ary_dup(ary));
+
+    if (NIL_P(v)) return ary;
+    return v;
 }
 
 static VALUE
 ary_nitems(ary)
-    struct RArray *ary;
+    VALUE ary;
 {
     int n = 0;
     VALUE *p, *pend;
 
-    p = ary->ptr;
-    pend = p + ary->len;
+    p = RARRAY(ary)->ptr;
+    pend = p + RARRAY(ary)->len;
     while (p < pend) {
 	if (!NIL_P(*p)) n++;
 	p++;
@@ -1171,7 +1283,34 @@ ary_nitems(ary)
     return INT2FIX(n);
 }
 
-extern VALUE mEnumerable;
+static VALUE
+ary_flatten_bang(ary)
+    VALUE ary;
+{
+    int i;
+    int mod = 0;
+
+    ary_modify(ary);
+    for (i=0; i<RARRAY(ary)->len; i++) {
+	VALUE ary2 = RARRAY(ary)->ptr[i];
+	if (TYPE(ary2) == T_ARRAY) {
+	    ary_replace(ary, i--, 1, ary2);
+	    mod = 1;
+	}
+    }
+    if (mod == 0) return Qnil;
+    return ary;
+}
+
+static VALUE
+ary_flatten(ary)
+    VALUE ary;
+{
+    VALUE v = ary_flatten_bang(ary_dup(ary));
+
+    if (NIL_P(v)) return ary;
+    return v;
+}
 
 void
 Init_Array()
@@ -1184,6 +1323,7 @@ Init_Array()
     rb_define_method(cArray, "to_s", ary_to_s, 0);
     rb_define_method(cArray, "inspect", ary_inspect, 0);
     rb_define_method(cArray, "to_a", ary_to_a, 0);
+    rb_define_method(cArray, "to_ary", ary_to_a, 0);
 
     rb_define_method(cArray, "freeze",  ary_freeze, 0);
     rb_define_method(cArray, "frozen?",  ary_frozen_p, 0);
@@ -1207,8 +1347,11 @@ Init_Array()
     rb_define_alias(cArray,  "size", "length");
     rb_define_method(cArray, "empty?", ary_empty_p, 0);
     rb_define_method(cArray, "index", ary_index, 1);
-    rb_define_method(cArray, "indexes", ary_indexes, -2);
+    rb_define_method(cArray, "rindex", ary_rindex, 1);
+    rb_define_method(cArray, "indexes", ary_indexes, -1);
+    rb_define_method(cArray, "indices", ary_indexes, -1);
     rb_define_method(cArray, "clone", ary_clone, 0);
+    rb_define_method(cArray, "dup", ary_dup, 0);
     rb_define_method(cArray, "join", ary_join_method, -1);
     rb_define_method(cArray, "reverse", ary_reverse_method, 0);
     rb_define_method(cArray, "reverse!", ary_reverse, 0);
@@ -1217,12 +1360,13 @@ Init_Array()
     rb_define_method(cArray, "delete", ary_delete, 1);
     rb_define_method(cArray, "delete_at", ary_delete_at, 1);
     rb_define_method(cArray, "delete_if", ary_delete_if, 0);
-#if 0
-    rb_define_method(cArray, "replace", ary_replace, 0);
-#endif
+    rb_define_method(cArray, "filter", ary_filter, 0);
+    rb_define_method(cArray, "replace", ary_replace_method, 1);
     rb_define_method(cArray, "clear", ary_clear, 0);
     rb_define_method(cArray, "fill", ary_fill, -1);
     rb_define_method(cArray, "include?", ary_includes, 1);
+    rb_define_method(cArray, "===", ary_includes, 1);
+    rb_define_method(cArray, "<=>", ary_cmp, 1);
 
     rb_define_method(cArray, "assoc", ary_assoc, 1);
     rb_define_method(cArray, "rassoc", ary_rassoc, 1);
@@ -1234,8 +1378,12 @@ Init_Array()
     rb_define_method(cArray, "&", ary_and, 1);
     rb_define_method(cArray, "|", ary_or, 1);
 
+    rb_define_method(cArray, "uniq", ary_uniq, 0);
+    rb_define_method(cArray, "uniq!", ary_uniq_bang, 0);
     rb_define_method(cArray, "compact", ary_compact, 0);
     rb_define_method(cArray, "compact!", ary_compact_bang, 0);
+    rb_define_method(cArray, "flatten", ary_flatten, 0);
+    rb_define_method(cArray, "flatten!", ary_flatten_bang, 0);
     rb_define_method(cArray, "nitems", ary_nitems, 0);
 
     cmp = rb_intern("<=>");
