@@ -2,7 +2,7 @@
  * ext/curses/curses.c
  * 
  * by MAEDA Shugo (ender@pic-internet.or.jp)
- * modified by Yukihiro Matsumoto (matz@ruby.club.or.jp)
+ * modified by Yukihiro Matsumoto (matz@netlab.co.jp)
  */
 
 #ifdef HAVE_NCURSES_H
@@ -12,12 +12,18 @@
 #  include <ncurses/curses.h>
 # else
 #  include <curses.h>
-#  if defined(__NetBSD__) && !defined(_maxx)
+#  if (defined(__bsdi__) || defined(__NetBSD__)) && !defined(_maxx)
 #   define _maxx maxx
 #  endif
-#  if defined(__NetBSD__) && !defined(_maxy)
+#  if (defined(__bsdi__) || defined(__NetBSD__)) && !defined(_maxy)
 #   define _maxy maxy
 #  endif
+#  if (defined(__bsdi__) || defined(__NetBSD__)) && !defined(_begx)
+#   define _begx begx
+#  endif
+#  if (defined(__bsdi__) || defined(__NetBSD__)) && !defined(_begy)
+#   define _begy begy
+#  endif
 # endif
 #endif
 
@@ -32,9 +38,6 @@ struct windata {
     WINDOW *window;
 };
 
-#define NUM2CHAR(x) (char)NUM2INT(x)
-#define CHAR2FIX(x) INT2FIX((int)x)
-
 static void
 no_window()
 {
@@ -46,23 +49,21 @@ no_window()
     if (winp->window == 0) no_window();\
 }
 
-static void
-curses_err()
-{
-    Fail("curses error");
-}
-
-#define CHECK(c) if ((c)==ERR) {curses_err();}
+#define CHECK(c) c
 
 static void
-free_window(struct windata *winp)
+free_window(winp)
+    struct windata *winp;
 {
     if (winp->window && winp->window != stdscr) delwin(winp->window);
     winp->window = 0;
+    free(winp);
 }
 
 static VALUE
-prep_window(VALUE class, WINDOW *window)
+prep_window(class, window)
+    VALUE class;
+    WINDOW *window;
 {
     VALUE obj;
     struct windata *winp;
@@ -96,7 +97,7 @@ curses_init_screen()
 static VALUE
 curses_stdscr()
 {
-    if (!rb_stdscr) curses_init_screen();
+    if (rb_stdscr == 0) curses_init_screen();
     return rb_stdscr;
 }
 
@@ -104,15 +105,29 @@ curses_stdscr()
 static VALUE
 curses_close_screen()
 {
-    CHECK(endwin());
+#ifdef HAVE_ISENDWIN
+    if (!isendwin())
+#endif
+	endwin();
     return Qnil;
 }
 
+static void
+curses_finalize()
+{
+    if (stdscr
+#ifdef HAVE_ISENDWIN
+	&& !isendwin()
+#endif
+	)
+	endwin();
+}
+
 /* def closed? */
 static VALUE
 curses_closed()
 {
-#ifdef HAVE_ENDWIN
+#ifdef HAVE_ISENDWIN
     if (isendwin()) {
 	return TRUE;
     }
@@ -124,7 +139,8 @@ curses_closed()
 
 /* def clear */
 static VALUE
-curses_clear(VALUE obj)
+curses_clear(obj)
+    VALUE obj;
 {
     wclear(stdscr);
     return Qnil;
@@ -132,87 +148,102 @@ curses_clear(VALUE obj)
 
 /* def refresh */
 static VALUE
-curses_refresh(VALUE obj)
+curses_refresh(obj)
+    VALUE obj;
 {
-    CHECK(refresh());
+    refresh();
     return Qnil;
 }
 
-/* def refresh */
+/* def doupdate */
 static VALUE
-curses_doupdate(VALUE obj)
+curses_doupdate(obj)
+    VALUE obj;
 {
-    CHECK(doupdate());
+#ifdef HAVE_DOUPDATE
+    doupdate();
+#else
+    refresh();
+#endif
     return Qnil;
 }
 
 /* def echo */
 static VALUE
-curses_echo(VALUE obj)
+curses_echo(obj)
+    VALUE obj;
 {
-    CHECK(echo());
+    echo();
     return Qnil;
 }
 
 /* def noecho */
 static VALUE
-curses_noecho(VALUE obj)
+curses_noecho(obj)
+    VALUE obj;
 {
-    CHECK(noecho());
+    noecho();
     return Qnil;
 }
 
 /* def raw */
 static VALUE
-curses_raw(VALUE obj)
+curses_raw(obj)
+    VALUE obj;
 {
-    CHECK(raw());
+    raw();
     return Qnil;
 }
 
 /* def noraw */
 static VALUE
-curses_noraw(VALUE obj)
+curses_noraw(obj)
+    VALUE obj;
 {
-    CHECK(noraw());
+    noraw();
     return Qnil;
 }
 
 /* def cbreak */
 static VALUE
-curses_cbreak(VALUE obj)
+curses_cbreak(obj)
+    VALUE obj;
 {
-    CHECK(cbreak());
+    cbreak();
     return Qnil;
 }
 
 /* def nocbreak */
 static VALUE
-curses_nocbreak(VALUE obj)
+curses_nocbreak(obj)
+    VALUE obj;
 {
-    CHECK(nocbreak());
+    nocbreak();
     return Qnil;
 }
 
 /* def nl */
 static VALUE
-curses_nl(VALUE obj)
+curses_nl(obj)
+    VALUE obj;
 {
-    CHECK(nl());
+    nl();
     return Qnil;
 }
 
 /* def nonl */
 static VALUE
-curses_nonl(VALUE obj)
+curses_nonl(obj)
+    VALUE obj;
 {
-    CHECK(nonl());
+    nonl();
     return Qnil;
 }
 
 /* def beep */
 static VALUE
-curses_beep(VALUE obj)
+curses_beep(obj)
+    VALUE obj;
 {
 #ifdef HAVE_BEEP
     beep();
@@ -222,18 +253,23 @@ curses_beep(VALUE obj)
 
 /* def flash */
 static VALUE
-curses_flash(VALUE obj)
+curses_flash(obj)
+    VALUE obj;
 {
+#ifdef HAVE_FLASH
     flash();
+#endif
     return Qnil;
 }
 
 /* def ungetch */
 static VALUE
-curses_ungetch(VALUE obj, VALUE ch)
+curses_ungetch(obj, ch)
+    VALUE obj;
+    VALUE ch;
 {
 #ifdef HAVE_UNGETCH
-    CHECK(ungetch(NUM2INT(ch)));
+    ungetch(NUM2INT(ch));
 #else
     rb_notimplement();
 #endif
@@ -242,15 +278,19 @@ curses_ungetch(VALUE obj, VALUE ch)
 
 /* def setpos(y, x) */
 static VALUE
-curses_setpos(VALUE obj, VALUE y, VALUE x)
+curses_setpos(obj, y, x)
+    VALUE obj;
+    VALUE y;
+    VALUE x;
 {
-    CHECK(move(NUM2INT(y), NUM2INT(x)));
+    move(NUM2INT(y), NUM2INT(x));
     return Qnil;
 }
 
 /* def standout */
 static VALUE
-curses_standout(VALUE obj)
+curses_standout(obj)
+    VALUE obj;
 {
     standout();
     return Qnil;
@@ -258,7 +298,8 @@ curses_standout(VALUE obj)
 
 /* def standend */
 static VALUE
-curses_standend(VALUE obj)
+curses_standend(obj)
+    VALUE obj;
 {
     standend();
     return Qnil;
@@ -266,64 +307,77 @@ curses_standend(VALUE obj)
 
 /* def inch */
 static VALUE
-curses_inch(VALUE obj)
+curses_inch(obj)
+    VALUE obj;
 {
-    return CHAR2FIX(inch());
+    return CHR2FIX(inch());
 }
 
 /* def addch(ch) */
 static VALUE
-curses_addch(VALUE obj, VALUE ch)
+curses_addch(obj, ch)
+    VALUE obj;
+    VALUE ch;
 {
-    CHECK(addch(NUM2CHAR(ch)));
+    addch(NUM2CHR(ch));
     return Qnil;
 }
 
 /* def insch(ch) */
 static VALUE
-curses_insch(VALUE obj, VALUE ch)
+curses_insch(obj, ch)
+    VALUE obj;
+    VALUE ch;
 {
-    CHECK(insch(NUM2CHAR(ch)));
+    insch(NUM2CHR(ch));
     return Qnil;
 }
 
 /* def addstr(str) */
 static VALUE
-curses_addstr(VALUE obj, VALUE str)
+curses_addstr(obj, str)
+    VALUE obj;
+    VALUE str;
 {
-    addstr(RSTRING(str)->ptr);
+    if (!NIL_P(str)) {
+	addstr(STR2CSTR(str));
+    }
     return Qnil;
 }
 
 /* def getch */
 static VALUE
-curses_getch(VALUE obj)
+curses_getch(obj)
+    VALUE obj;
 {
-    return CHAR2FIX(getch());
+    return CHR2FIX(getch());
 }
 
 /* def getstr */
 static VALUE
-curses_getstr(VALUE obj)
+curses_getstr(obj)
+    VALUE obj;
 {
     char rtn[1024]; /* This should be big enough.. I hope */
-    CHECK(getstr(rtn));
+    getstr(rtn);
     return str_taint(str_new2(rtn));
 }
 
 /* def delch */
 static VALUE
-curses_delch(VALUE obj)
+curses_delch(obj)
+    VALUE obj;
 {
-    CHECK(delch());
+    delch();
     return Qnil;
 }
 
 /* def delelteln */
 static VALUE
-curses_deleteln(VALUE obj)
+curses_deleteln(obj)
+    VALUE obj;
 {
-    CHECK(deleteln());
+    deleteln();
     return Qnil;
 }
 
@@ -343,22 +397,32 @@ curses_cols()
 
 /* def new(lines, cols, top, left) */
 static VALUE
-window_s_new(VALUE class,
-	     VALUE lines, VALUE cols,
-	     VALUE top, VALUE left)
+window_s_new(class, lines, cols, top, left)
+    VALUE class;
+    VALUE lines;
+    VALUE cols;
+    VALUE top;
+    VALUE left;
 {
+    VALUE w;
     WINDOW *window;
     
     window = newwin(NUM2INT(lines), NUM2INT(cols), NUM2INT(top), NUM2INT(left));
     wclear(window);
-    return prep_window(class, window);
+    w = prep_window(class, window);
+    obj_call_init(w);
+
+    return w;
 }
 
 /* def subwin(lines, cols, top, left) */
 static VALUE
-window_subwin(VALUE obj,
-	     VALUE lines, VALUE cols,
-	     VALUE top, VALUE left)
+window_subwin(obj, lines, cols, top, left)
+    VALUE obj;
+    VALUE lines;
+    VALUE cols;
+    VALUE top;
+    VALUE left;
 {
     struct windata *winp;
     WINDOW *window;
@@ -371,19 +435,22 @@ window_subwin(VALUE obj,
 
 /* def close */
 static VALUE
-window_close(VALUE obj)
+window_close(obj)
+    VALUE obj;
 {
     struct windata *winp;
     
     GetWINDOW(obj, winp);
-    free_window(winp);
+    delwin(winp->window);
+    winp->window = 0;
 
     return Qnil;
 }
 
 /* def clear */
 static VALUE
-window_clear(VALUE obj)
+window_clear(obj)
+    VALUE obj;
 {
     struct windata *winp;
     
@@ -395,24 +462,28 @@ window_clear(VALUE obj)
 
 /* def refresh */
 static VALUE
-window_refresh(VALUE obj)
+window_refresh(obj)
+    VALUE obj;
 {
     struct windata *winp;
     
     GetWINDOW(obj, winp);
-    CHECK(wrefresh(winp->window));
+    wrefresh(winp->window);
     
     return Qnil;
 }
 
 /* def box(vert, hor) */
 static VALUE
-window_box(VALUE obj, VALUE vert, VALUE hor)
+window_box(obj, vert, hor)
+    VALUE obj;
+    VALUE vert;
+    VALUE hor;
 {
     struct windata *winp; 
    
     GetWINDOW(obj, winp);
-    box(winp->window, NUM2CHAR(vert), NUM2CHAR(hor));
+    box(winp->window, NUM2CHR(vert), NUM2CHR(hor));
     
     return Qnil;
 }
@@ -420,30 +491,37 @@ window_box(VALUE obj, VALUE vert, VALUE hor)
 
 /* def move(y, x) */
 static VALUE
-window_move(VALUE obj, VALUE y, VALUE x)
+window_move(obj, y, x)
+    VALUE obj;
+    VALUE y;
+    VALUE x;
 {
     struct windata *winp;
     
     GetWINDOW(obj, winp);
-    CHECK(mvwin(winp->window, NUM2INT(y), NUM2INT(x)));
+    mvwin(winp->window, NUM2INT(y), NUM2INT(x));
 
     return Qnil;
 }
 
 /* def setpos(y, x) */
 static VALUE
-window_setpos(VALUE obj, VALUE y, VALUE x)
+window_setpos(obj, y, x)
+    VALUE obj;
+    VALUE y;
+    VALUE x;
 {
     struct windata *winp;
     
     GetWINDOW(obj, winp);
-    CHECK(wmove(winp->window, NUM2INT(y), NUM2INT(x)));
+    wmove(winp->window, NUM2INT(y), NUM2INT(x));
     return Qnil;
 }
 
 /* def cury */
 static VALUE
-window_cury(VALUE obj)
+window_cury(obj)
+    VALUE obj;
 {
     struct windata *winp;
     int x, y;
@@ -455,7 +533,8 @@ window_cury(VALUE obj)
 
 /* def curx */
 static VALUE
-window_curx(VALUE obj)
+window_curx(obj)
+    VALUE obj;
 {
     struct windata *winp;
     int x, y;
@@ -467,7 +546,8 @@ window_curx(VALUE obj)
 
 /* def maxy */
 static VALUE
-window_maxy(VALUE obj)
+window_maxy(obj)
+    VALUE obj;
 {
     struct windata *winp;
     int x, y;
@@ -487,7 +567,8 @@ window_maxy(VALUE obj)
 
 /* def maxx */
 static VALUE
-window_maxx(VALUE obj)
+window_maxx(obj)
+    VALUE obj;
 {
     struct windata *winp;
     int x, y;
@@ -507,7 +588,8 @@ window_maxx(VALUE obj)
 
 /* def begy */
 static VALUE
-window_begy(VALUE obj)
+window_begy(obj)
+    VALUE obj;
 {
     struct windata *winp;
     int x, y;
@@ -523,7 +605,8 @@ window_begy(VALUE obj)
 
 /* def begx */
 static VALUE
-window_begx(VALUE obj)
+window_begx(obj)
+    VALUE obj;
 {
     struct windata *winp;
     int x, y;
@@ -539,7 +622,8 @@ window_begx(VALUE obj)
 
 /* def standout */
 static VALUE
-window_standout(VALUE obj)
+window_standout(obj)
+    VALUE obj;
 {
     struct windata *winp;
     
@@ -550,7 +634,8 @@ window_standout(VALUE obj)
 
 /* def standend */
 static VALUE
-window_standend(VALUE obj)
+window_standend(obj)
+    VALUE obj;
 {
     struct windata *winp;
     
@@ -561,53 +646,63 @@ window_standend(VALUE obj)
 
 /* def inch */
 static VALUE
-window_inch(VALUE obj)
+window_inch(obj)
+    VALUE obj;
 {
     struct windata *winp;
     
     GetWINDOW(obj, winp);
-    return CHAR2FIX(winch(winp->window));
+    return CHR2FIX(winch(winp->window));
 }
 
 /* def addch(ch) */
 static VALUE
-window_addch(VALUE obj, VALUE ch)
+window_addch(obj, ch)
+    VALUE obj;
+    VALUE ch;
 {
     struct windata *winp;
     
     GetWINDOW(obj, winp);
-    CHECK(waddch(winp->window, NUM2CHAR(ch)));
+    waddch(winp->window, NUM2CHR(ch));
     
     return Qnil;
 }
 
 /* def insch(ch) */
 static VALUE
-window_insch(VALUE obj, VALUE ch)
+window_insch(obj, ch)
+    VALUE obj;
+    VALUE ch;
 {
     struct windata *winp;
     
     GetWINDOW(obj, winp);
-    CHECK(winsch(winp->window, NUM2CHAR(ch)));
+    winsch(winp->window, NUM2CHR(ch));
     
     return Qnil;
 }
 
 /* def addstr(str) */
 static VALUE
-window_addstr(VALUE obj, VALUE str)
+window_addstr(obj, str)
+    VALUE obj;
+    VALUE str;
 {
-    struct windata *winp;
-    
-    GetWINDOW(obj, winp);
-    CHECK(waddstr(winp->window, RSTRING(str)->ptr));
-    
+    if (!NIL_P(str)) {
+	struct windata *winp;
+
+	GetWINDOW(obj, winp);
+	waddstr(winp->window, STR2CSTR(str));
+    }
     return Qnil;
 }
 
 /* def <<(str) */
 static VALUE
-window_addstr2(VALUE obj, VALUE str)
+window_addstr2(obj, str)
+    VALUE obj;
+    VALUE str;
 {
     window_addstr(obj, str);
     return obj;
@@ -615,45 +710,49 @@ window_addstr2(VALUE obj, VALUE str)
 
 /* def getch */
 static VALUE
-window_getch(VALUE obj)
+window_getch(obj)
+    VALUE obj;
 {
     struct windata *winp;
     
     GetWINDOW(obj, winp);
-    return CHAR2FIX(wgetch(winp->window));
+    return CHR2FIX(wgetch(winp->window));
 }
 
 /* def getstr */
 static VALUE
-window_getstr(VALUE obj)
+window_getstr(obj)
+    VALUE obj;
 {
     struct windata *winp;
     char rtn[1024]; /* This should be big enough.. I hope */
     
     GetWINDOW(obj, winp);
-    CHECK(wgetstr(winp->window, rtn));
+    wgetstr(winp->window, rtn);
     return str_taint(str_new2(rtn));
 }
 
 /* def delch */
 static VALUE
-window_delch(VALUE obj)
+window_delch(obj)
+    VALUE obj;
 {
     struct windata *winp;
     
     GetWINDOW(obj, winp);
-    CHECK(wdelch(winp->window));
+    wdelch(winp->window);
     return Qnil;
 }
 
 /* def delelteln */
 static VALUE
-window_deleteln(VALUE obj)
+window_deleteln(obj)
+    VALUE obj;
 {
     struct windata *winp;
     
     GetWINDOW(obj, winp);
-    CHECK(wdeleteln(winp->window));
+    wdeleteln(winp->window);
     return Qnil;
 }
 
@@ -722,4 +821,6 @@ Init_curses()
     rb_define_method(cWindow, "getstr", window_getstr, 0);
     rb_define_method(cWindow, "delch", window_delch, 0);
     rb_define_method(cWindow, "deleteln", window_deleteln, 0);
+
+    rb_set_end_proc(curses_finalize, 0);
 }
