@@ -3,15 +3,16 @@
   file.c -
 
   $Author: matz $
-  $Date: 1996/12/25 10:42:36 $
+  $Date: 1998/09/03 07:43:31 $
   created at: Mon Nov 15 12:24:34 JST 1993
 
-  Copyright (C) 1993-1996 Yukihiro Matsumoto
+  Copyright (C) 1993-1998 Yukihiro Matsumoto
 
 ************************************************/
 
 #include "ruby.h"
-#include "io.h"
+#include "rubyio.h"
+#include "rubysig.h"
 
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
@@ -42,28 +43,23 @@ struct timeval {
 #include <pwd.h>
 #endif
 
-#ifdef HAVE_STRING_H
-# include <string.h>
-#else
-char *strrchr();
+#ifndef HAVE_STRING_H
+char *strrchr _((char*,char));
 #endif
 
-#ifdef NT
+#include <sys/types.h>
 #include <sys/stat.h>
+
+#ifdef USE_CWGUSI
+ #include "macruby_missing.h"
+ extern int fileno(FILE *stream);
+ extern int utimes();
 #endif
 
-#ifndef NT
-char *strdup();
-char *getenv();
-#endif
-
-extern VALUE cIO;
 VALUE cFile;
 VALUE mFileTest;
 static VALUE sStat;
 
-VALUE time_new();
-
 VALUE
 file_open(fname, mode)
     char *fname, *mode;
@@ -76,15 +72,16 @@ file_open(fname, mode)
     fptr->mode = io_mode_flags(mode);
     fptr->f = rb_fopen(fname, mode);
     fptr->path = strdup(fname);
+    obj_call_init((VALUE)port);
 
     return (VALUE)port;
 }
 
 static VALUE
-file_s_open(argc, argv, class)
+file_s_open(argc, argv, klass)
     int argc;
     VALUE *argv;
-    VALUE class;
+    VALUE klass;
 {
     VALUE fname, vmode, file;
     char *mode;
@@ -92,15 +89,19 @@ file_s_open(argc, argv, class)
     rb_scan_args(argc, argv, "11", &fname, &vmode);
     Check_SafeStr(fname);
     if (!NIL_P(vmode)) {
-	Check_Type(vmode, T_STRING);
-	mode = RSTRING(vmode)->ptr;
+	mode = STR2CSTR(vmode);
     }
     else {
 	mode = "r";
     }
     file = file_open(RSTRING(fname)->ptr, mode);
 
-    RBASIC(file)->class = class;
+    RBASIC(file)->klass = klass;
+    obj_call_init(file);
+    if (iterator_p()) {
+	return rb_ensure(rb_yield, file, io_close, file);
+    }
+
     return file;
 }
 
@@ -122,8 +123,7 @@ file_reopen(argc, argv, file)
 
     Check_SafeStr(fname);
     if (!NIL_P(nmode)) {
-	Check_Type(nmode, T_STRING);
-	mode = RSTRING(nmode)->ptr;
+	mode = STR2CSTR(nmode);
     }
     else {
 	mode = "r";
@@ -155,13 +155,14 @@ file_reopen(argc, argv, file)
 }
 
 static int
-apply2files(func, args, arg)
+apply2files(func, vargs, arg)
     int (*func)();
-    struct RArray *args;
+    VALUE vargs;
     void *arg;
 {
     int i;
     VALUE path;
+    struct RArray *args = RARRAY(vargs);
 
     for (i=0; i<args->len; i++) {
 	Check_SafeStr(args->ptr[i]);
@@ -176,76 +177,6 @@ apply2files(func, args, arg)
     return args->len;
 }
 
-static VALUE
-file_tell(obj)
-    VALUE obj;
-{
-    OpenFile *fptr;
-    long pos;
-
-    GetOpenFile(obj, fptr);
-
-    pos = ftell(fptr->f);
-    if (ferror(fptr->f) != 0) rb_sys_fail(0);
-
-    return int2inum(pos);
-}
-
-static VALUE
-file_seek(obj, offset, ptrname)
-    VALUE obj, offset, ptrname;
-{
-    OpenFile *fptr;
-    long pos;
-
-    GetOpenFile(obj, fptr);
-
-    pos = fseek(fptr->f, NUM2INT(offset), NUM2INT(ptrname));
-    if (pos != 0) rb_sys_fail(0);
-    clearerr(fptr->f);
-
-    return obj;
-}
-
-static VALUE
-file_set_pos(obj, offset)
-    VALUE obj, offset;
-{
-    OpenFile *fptr;
-    long pos;
-
-    GetOpenFile(obj, fptr);
-    pos = fseek(fptr->f, NUM2INT(offset), 0);
-    if (pos != 0) rb_sys_fail(0);
-    clearerr(fptr->f);
-
-    return obj;
-}
-
-static VALUE
-file_rewind(obj)
-    VALUE obj;
-{
-    OpenFile *fptr;
-
-    GetOpenFile(obj, fptr);
-    if (fseek(fptr->f, 0L, 0) != 0) rb_sys_fail(0);
-    clearerr(fptr->f);
-
-    return obj;
-}
-
-static VALUE
-file_eof(obj)
-    VALUE obj;
-{
-    OpenFile *fptr;
-
-    GetOpenFile(obj, fptr);
-    if (feof(fptr->f) == 0) return FALSE;
-    return TRUE;
-}
-
 static VALUE
 file_path(obj)
     VALUE obj;
@@ -253,23 +184,17 @@ file_path(obj)
     OpenFile *fptr;
 
     GetOpenFile(obj, fptr);
+    if (fptr->path == NULL) return Qnil;
     return str_new2(fptr->path);
 }
 
-static VALUE
-file_isatty(obj)
-    VALUE obj;
-{
-    return FALSE;
-}
-
-#include <sys/types.h>
 #ifndef NT
-#include <sys/file.h>
+# ifndef USE_CWGUSI
+#  include <sys/file.h>
+# endif
 #else
 #include "missing/file.h"
 #endif
-#include <sys/stat.h>
 
 static VALUE
 stat_new(st)
@@ -304,22 +229,36 @@ stat_new(st)
 		      time_new(st->st_ctime, 0));
 }
 
+static int
+rb_stat(file, st)
+    VALUE file;
+    struct stat *st;
+{
+    OpenFile *fptr;
+
+    if (TYPE(file) == T_FILE) {
+	GetOpenFile(file, fptr);
+	return fstat(fileno(fptr->f), st);
+    }
+    Check_SafeStr(file);
+    return stat(RSTRING(file)->ptr, st);
+}
+
 static VALUE
 file_s_stat(obj, fname)
-    VALUE obj;
-    struct RString *fname;
+    VALUE obj, fname;
 {
     struct stat st;
 
     Check_SafeStr(fname);
-    if (stat(fname->ptr, &st) == -1) {
-	rb_sys_fail(fname->ptr);
+    if (stat(RSTRING(fname)->ptr, &st) == -1) {
+	rb_sys_fail(RSTRING(fname)->ptr);
     }
     return stat_new(&st);
 }
 
 static VALUE
-file_stat(obj)
+io_stat(obj)
     VALUE obj;
 {
     OpenFile *fptr;
@@ -334,19 +273,18 @@ file_stat(obj)
 
 static VALUE
 file_s_lstat(obj, fname)
-    VALUE obj;
-    struct RString *fname;
+    VALUE obj, fname;
 {
 #if !defined(MSDOS) && !defined(NT)
     struct stat st;
 
     Check_SafeStr(fname);
-    if (lstat(fname->ptr, &st) == -1) {
-	rb_sys_fail(fname->ptr);
+    if (lstat(RSTRING(fname)->ptr, &st) == -1) {
+	rb_sys_fail(RSTRING(fname)->ptr);
     }
     return stat_new(&st);
 #else
-        rb_notimplement();
+    rb_notimplement();
 #endif
 }
 
@@ -364,7 +302,7 @@ file_lstat(obj)
     }
     return stat_new(&st);
 #else
-        rb_notimplement();
+    rb_notimplement();
 #endif
 }
 
@@ -372,7 +310,7 @@ static int
 group_member(gid)
     GETGROUPS_T gid;
 {
-#ifndef NT
+#if !defined(NT) && !defined(USE_CWGUSI)
     if (getgid() ==  gid || getegid() == gid)
 	return TRUE;
 
@@ -424,19 +362,6 @@ eaccess(path, mode)
 	return 0;
     }
 
-#if defined(DJGPP)
-  {
-    int stat_mode = 0;
-    if (mode & X_OK)
-      stat_mode |= S_IXOTH;
-    if (mode & W_OK)
-      stat_mode |= S_IWOTH;
-    if (mode & R_OK)
-      stat_mode |= S_IROTH;
-    mode = stat_mode;
-  }
-#endif
-
   if (st.st_uid == euid)        /* owner */
     mode <<= 6;
   else if (group_member (st.st_gid))
@@ -452,8 +377,7 @@ eaccess(path, mode)
 
 static VALUE
 test_d(obj, fname)
-    VALUE obj;
-    struct RString *fname;
+    VALUE obj, fname;
 {
 #ifndef S_ISDIR
 #   define S_ISDIR(m) ((m & S_IFMT) == S_IFDIR)
@@ -461,16 +385,14 @@ test_d(obj, fname)
 
     struct stat st;
 
-    Check_SafeStr(fname);
-    if (stat(fname->ptr, &st) < 0) return FALSE;
+    if (rb_stat(fname, &st) < 0) return FALSE;
     if (S_ISDIR(st.st_mode)) return TRUE;
     return FALSE;
 }
 
 static VALUE
 test_p(obj, fname)
-    VALUE obj;
-    struct RString *fname;
+    VALUE obj, fname;
 {
 #ifdef S_IFIFO
 #  ifndef S_ISFIFO
@@ -479,8 +401,7 @@ test_p(obj, fname)
 
     struct stat st;
 
-    Check_SafeStr(fname);
-    if (stat(fname->ptr, &st) < 0) return FALSE;
+    if (rb_stat(fname, &st) < 0) return FALSE;
     if (S_ISFIFO(st.st_mode)) return TRUE;
 
 #endif
@@ -489,8 +410,7 @@ test_p(obj, fname)
 
 static VALUE
 test_l(obj, fname)
-    VALUE obj;
-    struct RString *fname;
+    VALUE obj, fname;
 {
 #ifndef S_ISLNK
 #  ifdef _S_ISLNK
@@ -510,7 +430,7 @@ test_l(obj, fname)
     struct stat st;
 
     Check_SafeStr(fname);
-    if (lstat(fname->ptr, &st) < 0) return FALSE;
+    if (lstat(RSTRING(fname)->ptr, &st) < 0) return FALSE;
     if (S_ISLNK(st.st_mode)) return TRUE;
 
 #endif
@@ -519,8 +439,7 @@ test_l(obj, fname)
 
 static VALUE
 test_S(obj, fname)
-    VALUE obj;
-    struct RString *fname;
+    VALUE obj, fname;
 {
 #ifndef S_ISSOCK
 #  ifdef _S_ISSOCK
@@ -539,8 +458,7 @@ test_S(obj, fname)
 #ifdef S_ISSOCK
     struct stat st;
 
-    Check_SafeStr(fname);
-    if (stat(fname->ptr, &st) < 0) return FALSE;
+    if (rb_stat(fname, &st) < 0) return FALSE;
     if (S_ISSOCK(st.st_mode)) return TRUE;
 
 #endif
@@ -549,8 +467,7 @@ test_S(obj, fname)
 
 static VALUE
 test_b(obj, fname)
-    VALUE obj;
-    struct RString *fname;
+    VALUE obj, fname;
 {
 #ifndef S_ISBLK
 #   ifdef S_IFBLK
@@ -563,8 +480,7 @@ test_b(obj, fname)
 #ifdef S_ISBLK
     struct stat st;
 
-    Check_SafeStr(fname);
-    if (stat(fname->ptr, &st) < 0) return FALSE;
+    if (rb_stat(fname, &st) < 0) return FALSE;
     if (S_ISBLK(st.st_mode)) return TRUE;
 
 #endif
@@ -573,8 +489,7 @@ test_b(obj, fname)
 
 static VALUE
 test_c(obj, fname)
-    VALUE obj;
-    struct RString *fname;
+    VALUE obj, fname;
 {
 #ifndef S_ISCHR
 #   define S_ISCHR(m) ((m & S_IFMT) == S_IFCHR)
@@ -582,8 +497,7 @@ test_c(obj, fname)
 
     struct stat st;
 
-    Check_SafeStr(fname);
-    if (stat(fname->ptr, &st) < 0) return FALSE;
+    if (rb_stat(fname, &st) < 0) return FALSE;
     if (S_ISBLK(st.st_mode)) return TRUE;
 
     return FALSE;
@@ -591,73 +505,65 @@ test_c(obj, fname)
 
 static VALUE
 test_e(obj, fname)
-    VALUE obj;
-    struct RString *fname;
+    VALUE obj, fname;
 {
     struct stat st;
 
-    Check_SafeStr(fname);
-    if (stat(fname->ptr, &st) < 0) return FALSE;
+    if (rb_stat(fname, &st) < 0) return FALSE;
     return TRUE;
 }
 
 static VALUE
 test_r(obj, fname)
-    VALUE obj;
-    struct RString *fname;
+    VALUE obj, fname;
 {
     Check_SafeStr(fname);
-    if (eaccess(fname->ptr, R_OK) < 0) return FALSE;
+    if (eaccess(RSTRING(fname)->ptr, R_OK) < 0) return FALSE;
     return TRUE;
 }
 
 static VALUE
 test_R(obj, fname)
-    VALUE obj;
-    struct RString *fname;
+    VALUE obj, fname;
 {
     Check_SafeStr(fname);
-    if (access(fname->ptr, R_OK) < 0) return FALSE;
+    if (access(RSTRING(fname)->ptr, R_OK) < 0) return FALSE;
     return TRUE;
 }
 
 static VALUE
 test_w(obj, fname)
-    VALUE obj;
-    struct RString *fname;
+    VALUE obj, fname;
 {
     Check_SafeStr(fname);
-    if (eaccess(fname->ptr, W_OK) < 0) return FALSE;
+    if (eaccess(RSTRING(fname)->ptr, W_OK) < 0) return FALSE;
     return TRUE;
 }
 
 static VALUE
 test_W(obj, fname)
-    VALUE obj;
-    struct RString *fname;
+    VALUE obj, fname;
 {
     Check_SafeStr(fname);
-    if (access(fname->ptr, W_OK) < 0) return FALSE;
+    if (access(RSTRING(fname)->ptr, W_OK) < 0) return FALSE;
     return TRUE;
 }
 
 static VALUE
 test_x(obj, fname)
-    VALUE obj;
-    struct RString *fname;
+    VALUE obj, fname;
 {
     Check_SafeStr(fname);
-    if (eaccess(fname->ptr, X_OK) < 0) return FALSE;
+    if (eaccess(RSTRING(fname)->ptr, X_OK) < 0) return FALSE;
     return TRUE;
 }
 
 static VALUE
 test_X(obj, fname)
-    VALUE obj;
-    struct RString *fname;
+    VALUE obj, fname;
 {
     Check_SafeStr(fname);
-    if (access(fname->ptr, X_OK) < 0) return FALSE;
+    if (access(RSTRING(fname)->ptr, X_OK) < 0) return FALSE;
     return TRUE;
 }
 
@@ -667,79 +573,67 @@ test_X(obj, fname)
 
 static VALUE
 test_f(obj, fname)
-    VALUE obj;
-    struct RString *fname;
+    VALUE obj, fname;
 {
     struct stat st;
 
-    Check_SafeStr(fname);
-    if (stat(fname->ptr, &st) < 0) return FALSE;
+    if (rb_stat(fname, &st) < 0) return FALSE;
     if (S_ISREG(st.st_mode)) return TRUE;
     return FALSE;
 }
 
 static VALUE
 test_z(obj, fname)
-    VALUE obj;
-    struct RString *fname;
+    VALUE obj, fname;
 {
     struct stat st;
 
-    Check_SafeStr(fname);
-    if (stat(fname->ptr, &st) < 0) return FALSE;
+    if (rb_stat(fname, &st) < 0) return FALSE;
     if (st.st_size == 0) return TRUE;
     return FALSE;
 }
 
 static VALUE
 test_s(obj, fname)
-    VALUE obj;
-    struct RString *fname;
+    VALUE obj, fname;
 {
     struct stat st;
 
-    Check_SafeStr(fname);
-    if (stat(fname->ptr, &st) < 0) return FALSE;
+    if (rb_stat(fname, &st) < 0) return FALSE;
     if (st.st_size == 0) return FALSE;
     return int2inum(st.st_size);
 }
 
 static VALUE
 test_owned(obj, fname)
-    VALUE obj;
-    struct RString *fname;
+    VALUE obj, fname;
 {
     struct stat st;
 
-    Check_SafeStr(fname);
-    if (stat(fname->ptr, &st) < 0) return FALSE;
+    if (rb_stat(fname, &st) < 0) return FALSE;
     if (st.st_uid == geteuid()) return TRUE;
     return FALSE;
 }
 
 static VALUE
 test_rowned(obj, fname)
-    VALUE obj;
-    struct RString *fname;
+    VALUE obj, fname;
 {
     struct stat st;
 
-    Check_SafeStr(fname);
-    if (stat(fname->ptr, &st) < 0) return FALSE;
+    if (rb_stat(fname, &st) < 0) return FALSE;
     if (st.st_uid == getuid()) return TRUE;
     return FALSE;
 }
 
 static VALUE
 test_grpowned(obj, fname)
-    VALUE obj;
-    struct RString *fname;
+    VALUE obj, fname;
 {
 #ifndef NT
     struct stat st;
 
-    Check_SafeStr(fname);
-    if (stat(fname->ptr, &st) < 0) return FALSE;
+    if (rb_stat(fname, &st) < 0) return FALSE;
     if (st.st_gid == getegid()) return TRUE;
 #endif
     return FALSE;
@@ -761,12 +655,11 @@ check3rdbyte(file, mode)
 
 static VALUE
 test_suid(obj, fname)
-    VALUE obj;
-    struct RString *fname;
+    VALUE obj, fname;
 {
 #ifdef S_ISUID
     Check_SafeStr(fname);
-    return check3rdbyte(fname->ptr, S_ISUID);
+    return check3rdbyte(RSTRING(fname)->ptr, S_ISUID);
 #else
     return FALSE;
 #endif
@@ -774,12 +667,11 @@ test_suid(obj, fname)
 
 static VALUE
 test_sgid(obj, fname)
-    VALUE obj;
-    struct RString *fname;
+    VALUE obj, fname;
 {
 #ifndef NT
     Check_SafeStr(fname);
-    return check3rdbyte(fname->ptr, S_ISGID);
+    return check3rdbyte(RSTRING(fname)->ptr, S_ISGID);
 #else
     return FALSE;
 #endif
@@ -787,12 +679,10 @@ test_sgid(obj, fname)
 
 static VALUE
 test_sticky(obj, fname)
-    VALUE obj;
-    struct RString *fname;
+    VALUE obj, fname;
 {
-    Check_Type(fname, T_STRING);
 #ifdef S_ISVTX
-    return check3rdbyte(fname->ptr, S_ISVTX);
+    return check3rdbyte(STR2CSTR(fname), S_ISVTX);
 #else
     return FALSE;
 #endif
@@ -800,26 +690,31 @@ test_sticky(obj, fname)
 
 static VALUE
 file_s_size(obj, fname)
-    VALUE obj;
-    struct RString *fname;
+    VALUE obj, fname;
 {
     struct stat st;
 
-    Check_SafeStr(fname);
-    if (stat(fname->ptr, &st) < 0) rb_sys_fail(fname->ptr);
+    if (rb_stat(fname, &st) < 0)
+	rb_sys_fail(RSTRING(fname)->ptr);
     return int2inum(st.st_size);
 }
 
 static VALUE
 file_s_ftype(obj, fname)
-    VALUE obj;
-    struct RString *fname;
+    VALUE obj, fname;
 {
     struct stat st;
     char *t;
 
+#if defined(MSDOS) || defined(NT)
+    if (rb_stat(fname, &st) < 0)
+	rb_sys_fail(RSTRING(fname)->ptr);
+#else
     Check_SafeStr(fname);
-    if (stat(fname->ptr, &st) < 0) rb_sys_fail(fname->ptr);
+    if (lstat(RSTRING(fname)->ptr, &st) == -1) {
+	rb_sys_fail(RSTRING(fname)->ptr);
+    }
+#endif
 
     if (S_ISREG(st.st_mode)) {
 	t = "file";
@@ -857,13 +752,12 @@ file_s_ftype(obj, fname)
 
 static VALUE
 file_s_atime(obj, fname)
-    VALUE obj;
-    struct RString *fname;
+    VALUE obj, fname;
 {
     struct stat st;
 
-    Check_SafeStr(fname);
-    if (stat(fname->ptr, &st) < 0) rb_sys_fail(fname->ptr);
+    if (rb_stat(fname, &st) < 0)
+	rb_sys_fail(RSTRING(fname)->ptr);
     return time_new(st.st_atime, 0);
 }
 
@@ -883,13 +777,12 @@ file_atime(obj)
 
 static VALUE
 file_s_mtime(obj, fname)
-    VALUE obj;
-    struct RString *fname;
+    VALUE obj, fname;
 {
     struct stat st;
 
-    Check_SafeStr(fname);
-    if (stat(fname->ptr, &st) < 0) rb_sys_fail(fname->ptr);
+    if (rb_stat(fname, &st) < 0)
+	rb_sys_fail(RSTRING(fname)->ptr);
     return time_new(st.st_mtime, 0);
 }
 
@@ -909,13 +802,12 @@ file_mtime(obj)
 
 static VALUE
 file_s_ctime(obj, fname)
-    VALUE obj;
-    struct RString *fname;
+    VALUE obj, fname;
 {
     struct stat st;
 
-    Check_SafeStr(fname);
-    if (stat(fname->ptr, &st) < 0) rb_sys_fail(fname->ptr);
+    if (rb_stat(fname, &st) < 0)
+	rb_sys_fail(RSTRING(fname)->ptr);
     return time_new(st.st_ctime, 0);
 }
 
@@ -969,7 +861,7 @@ file_chmod(obj, vmode)
     mode = NUM2INT(vmode);
 
     GetOpenFile(obj, fptr);
-#if defined(DJGPP) || defined(__CYGWIN32__) || defined(NT)
+#if defined(DJGPP) || defined(NT) || defined(USE_CWGUSI) || defined(__BEOS__)
     if (chmod(fptr->path, mode) == -1)
 	rb_sys_fail(fptr->path);
 #else
@@ -1028,7 +920,7 @@ file_chown(obj, owner, group)
 
     rb_secure(2);
     GetOpenFile(obj, fptr);
-#if defined(DJGPP) || defined(__CYGWIN32__) || defined(NT)
+#if defined(DJGPP) || defined(__CYGWIN32__) || defined(NT) || defined(USE_CWGUSI)
     if (chown(fptr->path, NUM2INT(owner), NUM2INT(group)) == -1)
 	rb_sys_fail(fptr->path);
 #else
@@ -1074,8 +966,14 @@ file_s_utime(argc, argv)
 
 #ifndef HAVE_UTIME_H
 # ifdef NT
+#   if defined(__BORLANDC__)
+#     include <utime.h>
+#   else
 #  include <sys/utime.h>
+#   endif
+#   if defined(_MSC_VER)
 #  define utimbuf _utimbuf
+#   endif
 # else
 struct utimbuf {
     long actime;
@@ -1118,42 +1016,39 @@ file_s_utime(argc, argv)
 
 static VALUE
 file_s_link(obj, from, to)
-    VALUE obj;
-    struct RString *from, *to;
+    VALUE obj, from, to;
 {
-#ifndef __human68k__
+#if defined(USE_CWGUSI)
+        rb_notimplement();
+#else
     Check_SafeStr(from);
     Check_SafeStr(to);
 
-    if (link(from->ptr, to->ptr) < 0)
-	rb_sys_fail(from->ptr);
+    if (link(RSTRING(from)->ptr, RSTRING(to)->ptr) < 0)
+	rb_sys_fail(RSTRING(from)->ptr);
+    return INT2FIX(0);
+#endif /* USE_CWGUSI */
+}
+
+static VALUE
+file_s_symlink(obj, from, to)
+    VALUE obj, from, to;
+{
+#if !defined(MSDOS) && !defined(NT)
+    Check_SafeStr(from);
+    Check_SafeStr(to);
+
+    if (symlink(RSTRING(from)->ptr, RSTRING(to)->ptr) < 0)
+	rb_sys_fail(RSTRING(from)->ptr);
     return INT2FIX(0);
 #else
     rb_notimplement();
 #endif
 }
 
-static VALUE
-file_s_symlink(obj, from, to)
-    VALUE obj;
-    struct RString *from, *to;
-{
-#if !defined(MSDOS) && !defined(NT)
-    Check_SafeStr(from);
-    Check_SafeStr(to);
-
-    if (symlink(from->ptr, to->ptr) < 0)
-	rb_sys_fail(from->ptr);
-    return TRUE;
-#else
-        rb_notimplement();
-#endif
-}
-
 static VALUE
 file_s_readlink(obj, path)
-    VALUE obj;
-    struct RString *path;
+    VALUE obj, path;
 {
 #if !defined(MSDOS) && !defined(NT)
     char buf[MAXPATHLEN];
@@ -1161,8 +1056,8 @@ file_s_readlink(obj, path)
 
     Check_SafeStr(path);
 
-    if ((cc = readlink(path->ptr, buf, MAXPATHLEN)) < 0)
-	rb_sys_fail(path->ptr);
+    if ((cc = readlink(RSTRING(path)->ptr, buf, MAXPATHLEN)) < 0)
+	rb_sys_fail(RSTRING(path)->ptr);
 
     return str_new(buf, cc);
 #else
@@ -1180,8 +1075,7 @@ unlink_internal(path)
 
 static VALUE
 file_s_unlink(obj, args)
-    VALUE obj;
-    struct RArray *args;
+    VALUE obj, args;
 {
     int n;
 
@@ -1191,14 +1085,13 @@ file_s_unlink(obj, args)
 
 static VALUE
 file_s_rename(obj, from, to)
-    VALUE obj;
-    struct RString *from, *to;
+    VALUE obj, from, to;
 {
     Check_SafeStr(from);
     Check_SafeStr(to);
 
-    if (rename(from->ptr, to->ptr) == -1)
-	rb_sys_fail(from->ptr);
+    if (rename(RSTRING(from)->ptr, RSTRING(to)->ptr) < 0)
+	rb_sys_fail(RSTRING(from)->ptr);
 
     return INT2FIX(0);
 }
@@ -1208,10 +1101,13 @@ file_s_umask(argc, argv)
     int argc;
     VALUE *argv;
 {
+#ifdef USE_CWGUSI
+    rb_notimplement();
+#else
     int omask = 0;
 
     if (argc == 0) {
-	int omask = umask(0);
+	omask = umask(0);
 	umask(omask);
     }
     else if (argc == 1) {
@@ -1221,22 +1117,28 @@ file_s_umask(argc, argv)
 	ArgError("wrong # of argument");
     }
     return INT2FIX(omask);
+#endif /* USE_CWGUSI */
 }
 
 VALUE
-file_s_expand_path(obj, fname)
-    VALUE obj;
-    struct RString *fname;
+file_s_expand_path(argc, argv)
+    int argc;
+    VALUE *argv;
 {
+    VALUE fname, dname;
     char *s, *p;
-    char buf[MAXPATHLEN];
+    char buf[MAXPATHLEN+2];
 
-    Check_Type(fname, T_STRING);
-    s = fname->ptr;
+    rb_scan_args(argc, argv, "11", &fname, &dname);
 
+    s = STR2CSTR(fname);
     p = buf;
     if (s[0] == '~') {
-	if (s[1] == '/' || s[1] == '\0') {
+	if (s[1] == '/' || 
+#if defined(MSDOS) || defined(NT) || defined(__human68k__)
+	    s[1] == '\\' || 
+#endif
+	    s[1] == '\0') {
 	    char *dir = getenv("HOME");
 
 	    if (!dir) {
@@ -1251,7 +1153,6 @@ file_s_expand_path(obj, fname)
 	    struct passwd *pwPtr;
 	    s++;
 #endif
-
 	    while (*s && *s != '/') {
 		*p++ = *s++;
 	    }
@@ -1269,12 +1170,19 @@ file_s_expand_path(obj, fname)
 	}
     }
     else if (s[0] != '/') {
+	if (argc == 2) {
+	    dname = file_s_expand_path(1, &dname);
+	    strcpy(buf, RSTRING(dname)->ptr);
+	}
+	else {
 #ifdef HAVE_GETCWD
-	getcwd(buf, MAXPATHLEN);
+	    getcwd(buf, MAXPATHLEN);
 #else
-	getwd(buf);
+	    getwd(buf);
 #endif
+	}
 	p = &buf[strlen(buf)];
+	while (p > buf && *(p - 1) == '/') p--;
     }
     *p = '/';
 
@@ -1326,6 +1234,11 @@ rmext(p, e)
     if (!e) return 0;
 
     l2 = strlen(e);
+    if (l2 == 2 && e[1] == '*') {
+	e = strrchr(p, *e);
+	if (!e) return 0;
+	return e - p;
+    }
     if (l1 < l2) return l1;
 
     if (strcmp(p+l1-l2, e) == 0) {
@@ -1339,25 +1252,25 @@ file_s_basename(argc, argv)
     int argc;
     VALUE *argv;
 {
-    struct RString *fname;
-    struct RString *ext;
-    char *p;
+    VALUE fname, fext;
+    char *name, *p, *ext;
     int f;
 
-    rb_scan_args(argc, argv, "11", &fname, &ext);
-    Check_Type(fname, T_STRING);
-    if (!NIL_P(ext)) Check_Type(ext, T_STRING);
-    p = strrchr(fname->ptr, '/');
+    if (rb_scan_args(argc, argv, "11", &fname, &fext) == 2) {
+	ext = STR2CSTR(fext);
+    }
+    name = STR2CSTR(fname);
+    p = strrchr(name, '/');
     if (!p) {
-	if (!NIL_P(ext)) {
-	    f = rmext(fname->ptr, ext->ptr);
-	    if (f) return str_new(fname->ptr, f);
+	if (!NIL_P(fext)) {
+	    f = rmext(name, ext);
+	    if (f) return str_new(name, f);
 	}
-	return (VALUE)fname;
+	return fname;
     }
     p++;			/* skip last `/' */
-    if (!NIL_P(ext)) {
-	f = rmext(p, ext->ptr);
+    if (!NIL_P(fext)) {
+	f = rmext(p, ext);
 	if (f) return str_new(p, f);
     }
     return str_taint(str_new2(p));
@@ -1365,19 +1278,18 @@ file_s_basename(argc, argv)
 
 static VALUE
 file_s_dirname(obj, fname)
-    VALUE obj;
-    struct RString *fname;
+    VALUE obj, fname;
 {
-    UCHAR *p;
+    char *name, *p;
 
-    Check_Type(fname, T_STRING);
-    p = strrchr(fname->ptr, '/');
+    name = STR2CSTR(fname);
+    p = strrchr(name, '/');
     if (!p) {
 	return str_new2(".");
     }
-    if (p == fname->ptr)
+    if (p == name)
 	p++;
-    return str_taint(str_new(fname->ptr, p - fname->ptr));
+    return str_taint(str_new(name, p - name));
 }
 
 static VALUE
@@ -1398,31 +1310,30 @@ file_s_join(obj, args)
 
 static VALUE
 file_s_truncate(obj, path, len)
-    VALUE obj, len;
-    struct RString *path;
+    VALUE obj, path, len;
 {
     Check_SafeStr(path);
 
 #ifdef HAVE_TRUNCATE
-    if (truncate(path->ptr, NUM2INT(len)) < 0)
-	rb_sys_fail(path->ptr);
+    if (truncate(RSTRING(path)->ptr, NUM2INT(len)) < 0)
+	rb_sys_fail(RSTRING(path)->ptr);
 #else
 # ifdef HAVE_CHSIZE
     {
 	int tmpfd;
 
 #  if defined(NT)
-	if ((tmpfd = open(path->ptr, O_RDWR)) < 0) {
-	    rb_sys_fail(path->ptr);
+	if ((tmpfd = open(RSTRING(path)->ptr, O_RDWR)) < 0) {
+	    rb_sys_fail(RSTRING(path)->ptr);
 	}
 #  else
-	if ((tmpfd = open(path->ptr, 0)) < 0) {
-	    rb_sys_fail(path->ptr);
+	if ((tmpfd = open(RSTRING(path)->ptr, 0)) < 0) {
+	    rb_sys_fail(RSTRING(path)->ptr);
 	}
 #  endif
 	if (chsize(tmpfd, NUM2INT(len)) < 0) {
 	    close(tmpfd);
-	    rb_sys_fail(path->ptr);
+	    rb_sys_fail(RSTRING(path)->ptr);
 	}
 	close(tmpfd);
     }
@@ -1430,7 +1341,7 @@ file_s_truncate(obj, path, len)
     rb_notimplement();
 # endif
 #endif
-    return TRUE;
+    return INT2FIX(0);
 }
 
 static VALUE
@@ -1439,9 +1350,8 @@ file_truncate(obj, len)
 {
     OpenFile *fptr;
 
-    GetOpenFile(obj, fptr);
-
     rb_secure(2);
+    GetOpenFile(obj, fptr);
     if (!(fptr->mode & FMODE_WRITABLE)) {
 	Fail("not opened for writing");
     }
@@ -1456,19 +1366,59 @@ file_truncate(obj, len)
     rb_notimplement();
 # endif
 #endif
-    return TRUE;
+    return INT2FIX(0);
 }
 
+# ifndef LOCK_SH
+#  define LOCK_SH 1
+# endif
+# ifndef LOCK_EX
+#  define LOCK_EX 2
+# endif
+# ifndef LOCK_NB
+#  define LOCK_NB 4
+# endif
+# ifndef LOCK_UN
+#  define LOCK_UN 8
+# endif
+
+#if defined(THREAD) && defined(EWOULDBLOCK)
+static int
+thread_flock(fd, op)
+    int fd, op;
+{
+    if (thread_alone() || (op & LOCK_NB)) {
+	return flock(fd, op);
+    }
+    op |= LOCK_NB;
+    while (flock(fd, op) < 0) {
+	switch (errno) {
+	  case EINTR:		/* can be happen? */
+	  case EWOULDBLOCK:
+	    thread_schedule();	/* busy wait */
+	    break;
+	  default:
+	    return -1;
+	}
+    }
+    return 0;
+}
+#define flock thread_flock
+#endif
+
 static VALUE
 file_flock(obj, operation)
     VALUE obj;
     VALUE operation;
 {
+#ifdef USE_CWGUSI
+    rb_notimplement();
+#else
     OpenFile *fptr;
 
+    rb_secure(2);
     GetOpenFile(obj, fptr);
 
-    rb_secure(2);
     if (flock(fileno(fptr->f), NUM2INT(operation)) < 0) {
 #ifdef EWOULDBLOCK
 	if (errno == EWOULDBLOCK) {
@@ -1477,8 +1427,10 @@ file_flock(obj, operation)
 #endif
 	rb_sys_fail(fptr->path);
     }
-    return obj;
+    return INT2FIX(0);
+#endif /* USE_CWGUSI */
 }
+#undef flock
 
 static void
 test_check(n, argc, argv)
@@ -1490,7 +1442,16 @@ test_check(n, argc, argv)
     n+=1;
     if (n < argc) ArgError("Wrong # of arguments(%d for %d)", argc, n);
     for (i=1; i<n; i++) {
-	Check_SafeStr(argv[i]);
+	switch (TYPE(argv[i])) {
+	  case T_STRING:
+	    Check_SafeStr(argv[i]);
+	    break;
+	  case T_FILE:
+	    break;
+	  default:
+	    Check_Type(argv[i], T_STRING);
+	    break;
+	}
     }
 }
 
@@ -1504,12 +1465,7 @@ f_test(argc, argv)
     int cmd;
 
     if (argc == 0) ArgError("Wrong # of arguments");
-    if (TYPE(argv[0]) == T_STRING && RSTRING(argv[0])->len == 1) {
-	cmd = RSTRING(argv[0])->ptr[0];
-    }
-    else {
-	cmd = NUM2INT(argv[0]);
-    }
+    cmd = NUM2CHR(argv[0]);
     if (cmd == 0) return FALSE;
     if (strchr("bcdefgGkloOprRsSuwWxXz", cmd)) {
 	CHECK(1);
@@ -1587,7 +1543,7 @@ f_test(argc, argv)
 	struct stat st;
 
 	CHECK(1);
-	if (stat(RSTRING(argv[1])->ptr, &st) == -1) {
+	if (rb_stat(argv[1], &st) == -1) {
 	    rb_sys_fail(RSTRING(argv[1])->ptr);
 	}
 
@@ -1605,8 +1561,8 @@ f_test(argc, argv)
 	struct stat st1, st2;
 
 	CHECK(2);
-	if (stat(RSTRING(argv[1])->ptr, &st1) < 0) return FALSE;
-	if (stat(RSTRING(argv[2])->ptr, &st2) < 0) return FALSE;
+	if (rb_stat(argv[1], &st1) < 0) return FALSE;
+	if (rb_stat(argv[2], &st2) < 0) return FALSE;
 
 	switch (cmd) {
 	  case '-':
@@ -1627,14 +1583,16 @@ f_test(argc, argv)
 	    break;
 	}
     }
-    return FALSE;
+    /* unknown command */
+    ArgError("unknow command ?%c", cmd);
+    return Qnil;		/* not reached */
 }
 
-extern VALUE mKernel;
-
 void
 Init_File()
 {
+    VALUE mConst;
+
     mFileTest = rb_define_module("FileTest");
 
     rb_define_module_function(mFileTest, "directory?",  test_d, 1);
@@ -1649,6 +1607,7 @@ Init_File()
     rb_define_module_function(mFileTest, "file?",  test_f, 1);
     rb_define_module_function(mFileTest, "zero?",  test_z, 1);
     rb_define_module_function(mFileTest, "size?",  test_s, 1);
+    rb_define_module_function(mFileTest, "size",   test_s, 1);
     rb_define_module_function(mFileTest, "owned?",  test_owned, 1);
     rb_define_module_function(mFileTest, "grpowned?",  test_grpowned, 1);
 
@@ -1691,7 +1650,7 @@ Init_File()
     rb_define_singleton_method(cFile, "rename", file_s_rename, 2);
     rb_define_singleton_method(cFile, "umask", file_s_umask, -1);
     rb_define_singleton_method(cFile, "truncate", file_s_truncate, 2);
-    rb_define_singleton_method(cFile, "expand_path", file_s_expand_path, 1);
+    rb_define_singleton_method(cFile, "expand_path", file_s_expand_path, -1);
     rb_define_singleton_method(cFile, "basename", file_s_basename, -1);
     rb_define_singleton_method(cFile, "dirname", file_s_dirname, 1);
 
@@ -1702,7 +1661,7 @@ Init_File()
 
     rb_define_method(cFile, "reopen",  file_reopen, -1);
 
-    rb_define_method(cFile, "stat",  file_stat, 0);
+    rb_define_method(cIO, "stat",  io_stat, 0); /* this is IO's method */
     rb_define_method(cFile, "lstat",  file_lstat, 0);
 
     rb_define_method(cFile, "atime", file_atime, 0);
@@ -1713,38 +1672,19 @@ Init_File()
     rb_define_method(cFile, "chown", file_chown, 2);
     rb_define_method(cFile, "truncate", file_truncate, 1);
 
-    rb_define_method(cFile, "tell",  file_tell, 0);
-    rb_define_method(cFile, "seek",  file_seek, 2);
-
-    rb_define_method(cFile, "pos",  file_tell, 0);
-    rb_define_method(cFile, "pos=", file_set_pos, 1);
-
-    rb_define_method(cFile, "rewind", file_rewind, 0);
-    rb_define_method(cFile, "isatty", file_isatty, 0);
-    rb_define_method(cFile, "tty?",  file_isatty, 0);
-    rb_define_method(cFile, "eof", file_eof, 0);
-    rb_define_method(cFile, "eof?", file_eof, 0);
-
     rb_define_method(cFile, "flock", file_flock, 1);
 
-# ifndef LOCK_SH
-#  define LOCK_SH 1
-# endif
-# ifndef LOCK_EX
-#  define LOCK_EX 2
-# endif
-# ifndef LOCK_NB
-#  define LOCK_NB 4
-# endif
-# ifndef LOCK_UN
-#  define LOCK_UN 8
-# endif
-
+    mConst = rb_define_module_under(cFile, "Constants");
     rb_define_const(cFile, "LOCK_SH", INT2FIX(LOCK_SH));
     rb_define_const(cFile, "LOCK_EX", INT2FIX(LOCK_EX));
     rb_define_const(cFile, "LOCK_UN", INT2FIX(LOCK_UN));
     rb_define_const(cFile, "LOCK_NB", INT2FIX(LOCK_NB));
 
+    rb_define_const(mConst, "LOCK_SH", INT2FIX(LOCK_SH));
+    rb_define_const(mConst, "LOCK_EX", INT2FIX(LOCK_EX));
+    rb_define_const(mConst, "LOCK_UN", INT2FIX(LOCK_UN));
+    rb_define_const(mConst, "LOCK_NB", INT2FIX(LOCK_NB));
+
     rb_define_method(cFile, "path",  file_path, 0);
 
     rb_define_global_function("test", f_test, -1);
