@@ -3,15 +3,15 @@
   gc.c -
 
   $Author: matz $
-  $Date: 1996/12/25 09:54:47 $
+  $Date: 1998/12/25 04:35:41 $
   created at: Tue Oct  5 09:44:46 JST 1993
 
-  Copyright (C) 1993-1996 Yukihiro Matsumoto
+  Copyright (C) 1993-1998 Yukihiro Matsumoto
 
 ************************************************/
 
 #include "ruby.h"
-#include "sig.h"
+#include "rubysig.h"
 #include "st.h"
 #include "node.h"
 #include "env.h"
@@ -19,6 +19,9 @@
 #include <stdio.h>
 #include <setjmp.h>
 
+void re_free_registers _((struct re_registers*));
+void io_fptr_finalize _((struct OpenFile*));
+
 #ifndef setjmp
 #ifdef HAVE__SETJMP
 #define setjmp(env) _setjmp(env)
@@ -26,42 +29,43 @@
 #endif
 #endif
 
-#ifdef _AIX
-#pragma alloca
-#endif
-
 #ifdef C_ALLOCA
+#ifndef alloca
 void *alloca();
 #endif
+#endif
 
-void gc();
-void gc_mark();
 static void run_final();
 
 #ifndef GC_MALLOC_LIMIT
 #if defined(MSDOS) || defined(__human68k__)
-#define GC_MALLOC_LIMIT 200000
+#define GC_MALLOC_LIMIT 100000
 #else
-#define GC_MALLOC_LIMIT 400000
+#define GC_MALLOC_LIMIT 200000
 #endif
 #endif
+#define GC_NEWOBJ_LIMIT 1000
 
 static unsigned long malloc_memories = 0;
+static unsigned long alloc_objects = 0;
 
 void *
 xmalloc(size)
-    unsigned long size;
+    int size;
 {
     void *mem;
 
+    if (size < 0) {
+	ArgError("negative allocation size (or too big)");
+    }
     if (size == 0) size = 1;
     malloc_memories += size;
-    if (malloc_memories > GC_MALLOC_LIMIT) {
-	gc();
+    if (malloc_memories > GC_MALLOC_LIMIT && alloc_objects > GC_NEWOBJ_LIMIT) {
+	gc_gc();
     }
     mem = malloc(size);
     if (!mem) {
-	gc();
+	gc_gc();
 	mem = malloc(size);
 	if (!mem)
 	    Fatal("failed to allocate memory");
@@ -72,7 +76,7 @@ xmalloc(size)
 
 void *
 xcalloc(n, size)
-    unsigned long n, size;
+    int n, size;
 {
     void *mem;
 
@@ -85,14 +89,19 @@ xcalloc(n, size)
 void *
 xrealloc(ptr, size)
     void *ptr;
-    unsigned long size;
+    int size;
 {
     void *mem;
 
+    if (size < 0) {
+	ArgError("negative re-allocation size");
+    }
     if (!ptr) return xmalloc(size);
+    if (size == 0) size = 1;
+    malloc_memories += size;
     mem = realloc(ptr, size);
     if (!mem) {
-	gc();
+	gc_gc();
 	mem = realloc(ptr, size);
 	if (!mem)
 	    Fatal("failed to allocate memory(realloc)");
@@ -174,12 +183,12 @@ rb_global_variable(var)
 typedef struct RVALUE {
     union {
 	struct {
-	    UINT flag;		/* always 0 for freed obj */
+	    unsigned long flag;	/* always 0 for freed obj */
 	    struct RVALUE *next;
 	} free;
 	struct RBasic  basic;
 	struct RObject object;
-	struct RClass  class;
+	struct RClass  klass;
 	struct RFloat  flonum;
 	struct RString string;
 	struct RArray  array;
@@ -235,33 +244,35 @@ add_heap()
 	p++;
     }
 }
+#define RANY(o) ((RVALUE*)(o))
 
-struct RBasic *
+VALUE
 rb_newobj()
 {
-    struct RBasic *obj;
+    VALUE obj;
+
     if (freelist) {
       retry:
-	obj = (struct RBasic*)freelist;
+	obj = (VALUE)freelist;
 	freelist = freelist->as.free.next;
+	alloc_objects++;
 	return obj;
     }
-    if (dont_gc) add_heap();
-    else gc();
+    if (dont_gc || prohibit_interrupt) add_heap();
+    else gc_gc();
 
     goto retry;
 }
 
 VALUE
-data_object_alloc(class, datap, dmark, dfree)
-    VALUE class;
+data_object_alloc(klass, datap, dmark, dfree)
+    VALUE klass;
     void *datap;
     void (*dfree)();
     void (*dmark)();
 {
-    struct RData *data = (struct RData*)rb_newobj();
-
-    OBJSETUP(data, class, T_DATA);
+    NEWOBJ(data, struct RData);
+    OBJSETUP(data, klass, T_DATA);
     data->data = datap;
     data->dfree = dfree;
     data->dmark = dmark;
@@ -273,9 +284,10 @@ extern st_table *rb_class_tbl;
 VALUE *gc_stack_start;
 
 static int
-looks_pointerp(p)
-    register RVALUE *p;
+looks_pointerp(ptr)
+    void *ptr;
 {
+    register RVALUE *p = RANY(ptr);
     register RVALUE *heap_org;
     register long i;
 
@@ -365,14 +377,16 @@ gc_mark_maybe(obj)
 }
 
 void
-gc_mark(obj)
-    register RVALUE *obj;
+gc_mark(ptr)
+    void *ptr;
 {
+    register RVALUE *obj = RANY(ptr);
+
   Top:
     if (FIXNUM_P(obj)) return;	/* fixnum not marked */
     if (rb_special_const_p((VALUE)obj)) return; /* special const not marked */
     if (obj->as.basic.flags == 0) return; /* free cell */
-    if (obj->as.basic.flags & FL_MARK) return; /* marked */
+    if (obj->as.basic.flags & FL_MARK) return; /* already marked */
 
     obj->as.basic.flags |= FL_MARK;
 
@@ -387,6 +401,7 @@ gc_mark(obj)
 	  case NODE_IF:		/* 1,2,3 */
 	  case NODE_FOR:
 	  case NODE_ITER:
+	  case NODE_CREF:
 	    gc_mark(obj->as.node.u2.node);
 	    /* fall through */
 	  case NODE_BLOCK:	/* 1,3 */
@@ -403,16 +418,23 @@ gc_mark(obj)
 	  case NODE_SUPER:	/* 3 */
 	  case NODE_FCALL:
 	  case NODE_NEWLINE:
-	    obj = (RVALUE*)obj->as.node.u3.node;
+	    obj = RANY(obj->as.node.u3.node);
 	    goto Top;
 
 	  case NODE_WHILE:	/* 1,2 */
 	  case NODE_UNTIL:
+	  case NODE_MATCH2:
+	  case NODE_MATCH3:
 	    gc_mark(obj->as.node.u1.node);
 	    /* fall through */
 	  case NODE_METHOD:	/* 2 */
 	  case NODE_NOT:
-	    obj = (RVALUE*)obj->as.node.u2.node;
+	  case NODE_GASGN:
+	  case NODE_LASGN:
+	  case NODE_DASGN:
+	  case NODE_IASGN:
+	  case NODE_CASGN:
+	    obj = RANY(obj->as.node.u2.node);
 	    goto Top;
 
 	  case NODE_HASH:	/* 1 */
@@ -420,12 +442,13 @@ gc_mark(obj)
 	  case NODE_STR:
 	  case NODE_XSTR:
 	  case NODE_DEFINED:
-	    obj = (RVALUE*)obj->as.node.u1.node;
+	  case NODE_MATCH:
+	    obj = RANY(obj->as.node.u1.node);
 	    goto Top;
 
 	  case NODE_SCOPE:	/* 2,3 */
 	    gc_mark(obj->as.node.u3.node);
-	    obj = (RVALUE*)obj->as.node.u2.node;
+	    obj = RANY(obj->as.node.u2.node);
 	    goto Top;
 
 	  case NODE_ZARRAY:	/* - */
@@ -443,7 +466,15 @@ gc_mark(obj)
 	  case NODE_UNDEF:
 	  case NODE_SELF:
 	  case NODE_NIL:
+	  case NODE_POSTEXE:
 	    break;
+#ifdef C_ALLOCA
+	  case NODE_ALLOCA:
+	    mark_locations_array((VALUE*)obj->as.node.u1.value,
+				 obj->as.node.u3.cnt);
+	    obj = RANY(obj->as.node.u2.node);
+	    goto Top;
+#endif
 
 	  default:
 	    if (looks_pointerp(obj->as.node.u1.node)) {
@@ -453,26 +484,26 @@ gc_mark(obj)
 		gc_mark(obj->as.node.u2.node);
 	    }
 	    if (looks_pointerp(obj->as.node.u3.node)) {
-		obj = (RVALUE*)obj->as.node.u3.node;
+		obj = RANY(obj->as.node.u3.node);
 		goto Top;
 	    }
 	}
 	return;			/* no need to mark class. */
     }
 
-    gc_mark(obj->as.basic.class);
+    gc_mark(obj->as.basic.klass);
     switch (obj->as.basic.flags & T_MASK) {
       case T_ICLASS:
-	gc_mark(obj->as.class.super);
-	mark_tbl(obj->as.class.iv_tbl);
-	mark_tbl(obj->as.class.m_tbl);
+	gc_mark(obj->as.klass.super);
+	mark_tbl(obj->as.klass.iv_tbl);
+	mark_tbl(obj->as.klass.m_tbl);
 	break;
 
       case T_CLASS:
       case T_MODULE:
-	gc_mark(obj->as.class.super);
-	mark_tbl(obj->as.class.m_tbl);
-	mark_tbl(obj->as.class.iv_tbl);
+	gc_mark(obj->as.klass.super);
+	mark_tbl(obj->as.klass.m_tbl);
+	mark_tbl(obj->as.klass.iv_tbl);
 	break;
 
       case T_ARRAY:
@@ -491,7 +522,7 @@ gc_mark(obj)
 
       case T_STRING:
 	if (obj->as.string.orig) {
-	    obj = (RVALUE*)obj->as.string.orig;
+	    obj = RANY(obj->as.string.orig);
 	    goto Top;
 	}
 	break;
@@ -512,14 +543,14 @@ gc_mark(obj)
 
       case T_MATCH:
 	if (obj->as.match.str) {
-	    obj = (RVALUE*)obj->as.match.str;
+	    obj = RANY(obj->as.match.str);
 	    goto Top;
 	}
 	break;
 
       case T_VARMAP:
 	gc_mark(obj->as.varmap.val);
-	obj = (RVALUE*)obj->as.varmap.next;
+	obj = RANY(obj->as.varmap.next);
 	goto Top;
 	break;
 
@@ -584,7 +615,7 @@ gc_sweep()
 
 	while (p < pend) {
 	    if (!(p->as.basic.flags & FL_MARK)) {
-		if (p->as.basic.flags) obj_free(p);
+		if (p->as.basic.flags) obj_free((VALUE)p);
 		p->as.free.flag = 0;
 		p->as.free.next = nfreelist;
 		nfreelist = p;
@@ -597,6 +628,7 @@ gc_sweep()
 	freed += n;
 	freelist = nfreelist;
     }
+    alloc_objects = 0;
     if (freed < FREE_MIN) {
 	add_heap();
     }
@@ -604,20 +636,20 @@ gc_sweep()
 
 void
 gc_force_recycle(p)
-    RVALUE *p;
+    VALUE p;
 {
-    p->as.free.flag = 0;
-    p->as.free.next = freelist;
-    freelist = p;
+    RANY(p)->as.free.flag = 0;
+    RANY(p)->as.free.next = freelist;
+    freelist = RANY(p);
 }
 
 static int need_call_final = 0;
 
 static void
 obj_free(obj)
-    RVALUE *obj;
+    VALUE obj;
 {
-    switch (obj->as.basic.flags & T_MASK) {
+    switch (RANY(obj)->as.basic.flags & T_MASK) {
       case T_NIL:
       case T_FIXNUM:
       case T_TRUE:
@@ -629,40 +661,57 @@ obj_free(obj)
     if (need_call_final) {
 	run_final(obj);
     }
-    switch (obj->as.basic.flags & T_MASK) {
+    switch (RANY(obj)->as.basic.flags & T_MASK) {
       case T_OBJECT:
-	if (obj->as.object.iv_tbl) st_free_table(obj->as.object.iv_tbl);
+	if (RANY(obj)->as.object.iv_tbl) {
+	    st_free_table(RANY(obj)->as.object.iv_tbl);
+	}
 	break;
       case T_MODULE:
       case T_CLASS:
 	rb_clear_cache();
-	st_free_table(obj->as.class.m_tbl);
-	if (obj->as.object.iv_tbl) st_free_table(obj->as.object.iv_tbl);
+	st_free_table(RANY(obj)->as.klass.m_tbl);
+	if (RANY(obj)->as.object.iv_tbl) {
+	    st_free_table(RANY(obj)->as.object.iv_tbl);
+	}
 	break;
       case T_STRING:
-	if (!obj->as.string.orig) free(obj->as.string.ptr);
+#define STR_NO_ORIG FL_USER3	/* copied from string.c */
+	if (!RANY(obj)->as.string.orig || FL_TEST(obj, STR_NO_ORIG))
+	    free(RANY(obj)->as.string.ptr);
 	break;
       case T_ARRAY:
-	if (obj->as.array.ptr) free(obj->as.array.ptr);
+	if (RANY(obj)->as.array.ptr) free(RANY(obj)->as.array.ptr);
 	break;
       case T_HASH:
-	st_free_table(obj->as.hash.tbl);
+	if (RANY(obj)->as.hash.tbl)
+	    st_free_table(RANY(obj)->as.hash.tbl);
 	break;
       case T_REGEXP:
-	reg_free(obj->as.regexp.ptr);
-	free(obj->as.regexp.str);
+	if (RANY(obj)->as.regexp.ptr) re_free_pattern(RANY(obj)->as.regexp.ptr);
+	if (RANY(obj)->as.regexp.str) free(RANY(obj)->as.regexp.str);
 	break;
       case T_DATA:
-	if (obj->as.data.dfree && DATA_PTR(obj))
-	    (*obj->as.data.dfree)(DATA_PTR(obj));
+	if (DATA_PTR(obj)) {
+	    if ((long)RANY(obj)->as.data.dfree == -1) {
+		free(DATA_PTR(obj));
+	    }
+	    else if (RANY(obj)->as.data.dfree) {
+		(*RANY(obj)->as.data.dfree)(DATA_PTR(obj));
+	    }
+	}
 	break;
       case T_MATCH:
-	re_free_registers(obj->as.match.regs);
-	free(obj->as.match.regs);
+	if (RANY(obj)->as.match.regs)
+	    re_free_registers(RANY(obj)->as.match.regs);
+	if (RANY(obj)->as.match.regs)
+	    free(RANY(obj)->as.match.regs);
 	break;
       case T_FILE:
-	io_fptr_finalize(obj->as.file.fptr);
-	free(obj->as.file.fptr);
+	if (RANY(obj)->as.file.fptr) {
+	    io_fptr_finalize(RANY(obj)->as.file.fptr);
+	    free(RANY(obj)->as.file.fptr);
+	}
 	break;
       case T_ICLASS:
 	/* iClass shares table with the module */
@@ -673,30 +722,31 @@ obj_free(obj)
 	break;
 
       case T_BIGNUM:
-	if (obj->as.bignum.digits) free(obj->as.bignum.digits);
+	if (RANY(obj)->as.bignum.digits) free(RANY(obj)->as.bignum.digits);
 	break;
       case T_NODE:
-	if (nd_type(obj) == NODE_SCOPE && obj->as.node.u1.tbl) {
-	    free(obj->as.node.u1.tbl);
+	if (nd_type(obj) == NODE_SCOPE && RANY(obj)->as.node.u1.tbl) {
+	    free(RANY(obj)->as.node.u1.tbl);
 	}
 	return;			/* no need to free iv_tbl */
 
       case T_SCOPE:
-	if (obj->as.scope.local_vars) {
-	    VALUE *vars = obj->as.scope.local_vars-1;
+	if (RANY(obj)->as.scope.local_vars) {
+	    VALUE *vars = RANY(obj)->as.scope.local_vars-1;
 	    if (vars[0] == 0)
-		free(obj->as.scope.local_tbl);
-	    if (obj->as.scope.flag&SCOPE_MALLOC)
+		free(RANY(obj)->as.scope.local_tbl);
+	    if (RANY(obj)->as.scope.flag&SCOPE_MALLOC)
 		free(vars);
 	}
 	break;
 
       case T_STRUCT:
-	free(obj->as.rstruct.ptr);
+	if (RANY(obj)->as.rstruct.ptr)
+	    free(RANY(obj)->as.rstruct.ptr);
 	break;
 
       default:
-	Bug("gc_sweep(): unknown data type %d", obj->as.basic.flags & T_MASK);
+	Bug("gc_sweep(): unknown data type %d", RANY(obj)->as.basic.flags & T_MASK);
     }
 }
 
@@ -750,7 +800,7 @@ int rb_setjmp (rb_jmp_buf);
 #endif /* __GNUC__ */
 
 void
-gc()
+gc_gc()
 {
     struct gc_list *list;
     struct FRAME *frame;
@@ -806,7 +856,7 @@ gc()
 static VALUE
 gc_method()
 {
-    gc();
+    gc_gc();
     return Qnil;
 }
 
@@ -851,7 +901,7 @@ os_live_obj()
 		  case T_CLASS:
 		    if (FL_TEST(p, FL_SINGLETON)) continue;
 		  default:
-		    rb_yield(p);
+		    rb_yield((VALUE)p);
 		    n++;
 		}
 	    }
@@ -883,8 +933,8 @@ os_obj_of(of)
 		  case T_CLASS:
 		    if (FL_TEST(p, FL_SINGLETON)) continue;
 		  default:
-		    if (obj_is_kind_of(p, of)) {
-			rb_yield(p);
+		    if (obj_is_kind_of((VALUE)p, of)) {
+			rb_yield((VALUE)p);
 			n++;
 		    }
 		}
@@ -916,8 +966,6 @@ static VALUE
 add_final(os, proc)
     VALUE os, proc;
 {
-    extern VALUE cProc;
-
     if (!obj_is_kind_of(proc, cProc)) {
 	ArgError("wrong type argument %s (Proc required)",
 		  rb_class2name(CLASS_OF(proc)));
@@ -957,13 +1005,58 @@ run_final(obj)
 
     if (!FL_TEST(obj, FL_FINALIZE)) return;
 
-    obj |= FIXNUM_FLAG;		/* make obj into id */
+    obj = INT2NUM((long)obj);	/* make obj into id */
     for (i=0; i<RARRAY(finalizers)->len; i++) {
-	rb_eval_cmd(RARRAY(finalizers)->ptr[i], obj);
+	rb_eval_cmd(RARRAY(finalizers)->ptr[i], ary_new3(1,obj));
     }
 }
 
-extern VALUE cModule;
+void
+gc_call_finalizer_at_exit()
+{
+    RVALUE *p, *pend;
+    int i;
+
+    /* run finalizers */
+    for (i = 0; i < heaps_used; i++) {
+	p = heaps[i]; pend = p + HEAP_SLOTS;
+	while (p < pend) {
+	    run_final((VALUE)p);
+	    p++;
+	}
+    }
+    /* run data object's finaliers */
+    for (i = 0; i < heaps_used; i++) {
+	p = heaps[i]; pend = p + HEAP_SLOTS;
+	while (p < pend) {
+	    if (BUILTIN_TYPE(p) == T_DATA &&
+		DATA_PTR(p) &&
+		RANY(p)->as.data.dfree)
+		(*RANY(p)->as.data.dfree)(DATA_PTR(p));
+	    p++;
+	}
+    }
+}
+
+static VALUE
+id2ref(obj, id)
+    VALUE obj, id;
+{
+    unsigned long ptr = NUM2UINT(id);
+
+    if (FIXNUM_P(ptr)) return (VALUE)ptr;
+    if (ptr == TRUE) return TRUE;
+    if (ptr == FALSE) return FALSE;
+    if (ptr == Qnil) return Qnil;
+
+    if (!looks_pointerp(ptr)) {
+	IndexError("0x%x is not the id value", ptr);
+    }
+    if (RANY(ptr)->as.free.flag == 0) {
+	IndexError("0x%x is recycled object", ptr);
+    }
+    return (VALUE)ptr;
+}
 
 void
 Init_GC()
@@ -978,11 +1071,12 @@ Init_GC()
 
     mObSpace = rb_define_module("ObjectSpace");
     rb_define_module_function(mObSpace, "each_object", os_each_obj, -1);
-    rb_define_module_function(mObSpace, "garbage_collect", gc, 0);
+    rb_define_module_function(mObSpace, "garbage_collect", gc_method, 0);
     rb_define_module_function(mObSpace, "add_finalizer", add_final, 1);
     rb_define_module_function(mObSpace, "remove_finalizer", rm_final, 1);
     rb_define_module_function(mObSpace, "finalizers", finals, 0);
     rb_define_module_function(mObSpace, "call_finalizer", call_final, 1);
+    rb_define_module_function(mObSpace, "_id2ref", id2ref, 1);
 
     rb_global_variable(&finalizers);
     finalizers = ary_new();
