@@ -149,8 +149,8 @@ flock(int fd, int oper)
 #undef LK_LEN
 
 
-#undef const
-FILE *fdopen(int, const char *);
+//#undef const
+//FILE *fdopen(int, const char *);
 
 #if 0
 void
@@ -185,7 +185,7 @@ NtInitialize(int *argc, char ***argv) {
     tzset();
 
 	// Initialize Winsock
-	// StartSockets();
+	StartSockets();
 }
 
 
@@ -243,7 +243,7 @@ int SafeFree(char **vec, int vecc)
 
 	char *p;
 
-	p = (char *)(vec - (vecc * sizeof (char *) + 1));
+	p = (char *)vec;
 	free(p);
 
 	return 0;
@@ -251,6 +251,9 @@ int SafeFree(char **vec, int vecc)
 
 
 static char *szInternalCmds[] = {
+  "append",
+  "break",
+  "call",
   "cd",
   "chdir",
   "cls",
@@ -264,6 +267,7 @@ static char *szInternalCmds[] = {
   "md",
   "mkdir",
   "path",
+  "pause",
   "rd",
   "rem",
   "ren",
@@ -275,18 +279,25 @@ static char *szInternalCmds[] = {
   "type",
   "ver",
   "vol",
+  NULL
 };
 
 int
 isInternalCmd(char *cmd)
 {
-	int fRet;
+	int i, fRet=0;
 	char **vec;
-	int vecc = NtMakeCmdVector(cmd, &vec, FALSE);
-
+	NtMakeCmdVector(cmd, &vec, FALSE);
+        for( i = 0; szInternalCmds[i] ; i++){
+	    if(!strcmp(szInternalCmds[i], vec[0])){
+		fRet = 1;
+		break;
+	    }
+        }
+ 
 	SafeFree (vec, vecc);
 
-	return 0;
+	return fRet;
 }
 
 
@@ -501,7 +512,6 @@ mypopen (char *cmd, char *mode)
 				return fp;
 		}
 
-
 		fRet = CreatePipe(&hInFile, &hOutFile, &sa, 2048L);
 		if (!fRet)
 			Fatal("cannot open pipe \"%s\" (%s)", cmd, strerror(errno));
@@ -1572,7 +1582,7 @@ valid_filename(char *s)
 //
 
 FILE *
-fdopen (int fd, const char *mode)
+myfdopen (int fd, const char *mode)
 {
     FILE *fp;
     char sockbuf[80];
@@ -1580,6 +1590,8 @@ fdopen (int fd, const char *mode)
     int retval;
     extern int errno;
 
+    //fprintf(stderr, "myfdopen()\n");
+
     retval = getsockopt((SOCKET)fd, SOL_SOCKET, SO_TYPE, sockbuf, &optlen);
     if (retval == SOCKET_ERROR) {
 	int iRet;
@@ -1951,8 +1963,10 @@ mysocket (int af, int type, int protocol)
     if (!NtSocketsInitialized++) {
 	StartSockets();
     }
-    if ((s = socket (af, type, protocol)) == INVALID_SOCKET)
+    if ((s = socket (af, type, protocol)) == INVALID_SOCKET) {
 	errno = WSAGetLastError();
+	//fprintf(stderr, "socket fail (%d)", WSAGetLastError());
+    }
     return s;
 }
 
