@@ -3,22 +3,31 @@
   signal.c -
 
   $Author: matz $
-  $Date: 1996/12/25 10:42:53 $
+  $Date: 1998/07/09 08:40:26 $
   created at: Tue Dec 20 10:13:44 JST 1994
 
 ************************************************/
 
 #include "ruby.h"
-#include "sig.h"
+#include "rubysig.h"
 #include <signal.h>
 #include <stdio.h>
 
-#ifndef NSIG
-#ifdef DJGPP
-#define NSIG SIGMAX
-#else
-#define NSIG (_SIGMAX + 1)      /* For QNX */
+#ifdef __BEOS__
+#undef SIGBUS
 #endif
+
+#ifndef NSIG
+# ifdef DJGPP
+#  define NSIG SIGMAX
+# else
+#  define NSIG (_SIGMAX + 1)      /* For QNX */
+# endif
+#endif
+
+#ifdef USE_CWGUSI
+# undef NSIG
+# define NSIG __signal_max
 #endif
 
 static struct signals {
@@ -175,6 +184,10 @@ f_kill(argc, argv)
     int argc;
     VALUE *argv;
 {
+#ifdef USE_CWGUSI
+    rb_notimplement();
+#else
+    int negative = 0;
     int sig;
     int i;
     char *s;
@@ -193,9 +206,7 @@ f_kill(argc, argv)
 	break;
 
       case T_STRING:
-	{
-	    int negative = 0;
-
+        {
 	    s = RSTRING(argv[0])->ptr;
 	    if (s[0] == '-') {
 		negative++;
@@ -237,6 +248,7 @@ f_kill(argc, argv)
 	}
     }
     return INT2FIX(i-1);
+#endif /* USE_CWGUSI */
 }
 
 static VALUE trap_list[NSIG];
@@ -248,12 +260,14 @@ int prohibit_interrupt;
 void
 gc_mark_trap_list()
 {
+#ifndef MACOS_UNUSE_SIGNAL
     int i;
 
     for (i=0; i<NSIG; i++) {
 	if (trap_list[i])
 	    gc_mark(trap_list[i]);
     }
+#endif /* MACOS_UNUSE_SIGNAL */
 }
 
 #ifdef POSIX_SIGNAL
@@ -271,6 +285,11 @@ posix_signal(signum, handler)
 }
 #endif
 
+#ifdef THREAD
+# define rb_interrupt thread_interrupt
+# define rb_trap_eval thread_trap_eval
+#endif
+
 static RETSIGTYPE
 sighandle(sig)
     int sig;
@@ -285,11 +304,7 @@ sighandle(sig)
     if (trap_immediate) {
 	trap_immediate = 0;
 	if (sig == SIGINT && !trap_list[SIGINT]) {
-#ifdef THREAD
-	    thread_interrupt();
-#else
 	    rb_interrupt();
-#endif
 	}
 	rb_trap_eval(trap_list[sig], sig);
 	trap_immediate = 1;
@@ -321,34 +336,34 @@ sigsegv(sig)
 void
 rb_trap_exit()
 {
+#ifndef MACOS_UNUSE_SIGNAL
     if (trap_list[0])
 	rb_eval_cmd(trap_list[0], ary_new3(1, INT2FIX(0)));
+#endif
 }
 
 void
 rb_trap_exec()
 {
+#ifndef MACOS_UNUSE_SIGNAL
     int i;
 
     for (i=0; i<NSIG; i++) {
 	if (trap_pending_list[i]) {
 	    trap_pending_list[i] = 0;
 	    if (i == SIGINT && trap_list[SIGINT] == 0) {
-#ifdef THREAD
-		thread_interrupt();
-#else
 		rb_interrupt();
-#endif
 		return;
 	    }
 	    rb_trap_eval(trap_list[i], i);
 	}
     }
+#endif /* MACOS_UNUSE_SIGNAL */
     trap_pending = 0;
 }
 
 struct trap_arg {
-#ifndef NT
+#if !defined(NT) && !defined(USE_CWGUSI)
 # ifdef HAVE_SIGPROCMASK
     sigset_t mask;
 # else
@@ -358,6 +373,12 @@ struct trap_arg {
     VALUE sig, cmd;
 };
 
+# ifdef HAVE_SIGPROCMASK
+static sigset_t trap_last_mask;
+# else
+static int trap_last_mask;
+# endif
+
 static RETSIGTYPE
 sigexit()
 {
@@ -455,7 +476,7 @@ trap(arg)
 
     trap_list[sig] = command;
     /* enable at least specified signal. */
-#ifndef NT
+#if !defined(NT) && !defined(USE_CWGUSI)
 #ifdef HAVE_SIGPROCMASK
     sigdelset(&arg->mask, sig);
 #else
@@ -465,8 +486,8 @@ trap(arg)
     return old;
 }
 
-#ifndef NT
-static void
+#if !defined(NT) && !defined(USE_CWGUSI)
+static VALUE
 trap_ensure(arg)
     struct trap_arg *arg;
 {
@@ -476,9 +497,23 @@ trap_ensure(arg)
 #else
     sigsetmask(arg->mask);
 #endif
+    trap_last_mask = arg->mask;
+    return 0;
 }
 #endif
 
+void
+trap_restore_mask()
+{
+#ifndef NT
+# ifdef HAVE_SIGPROCMASK
+    sigprocmask(SIG_SETMASK, &trap_last_mask, NULL);
+# else
+    sigsetmask(trap_last_mask);
+# endif
+#endif
+}
+
 static VALUE
 f_trap(argc, argv)
     int argc;
@@ -499,7 +534,7 @@ f_trap(argc, argv)
 	arg.cmd = argv[1];
     }
 
-#ifndef NT
+#if !defined(NT) && !defined(USE_CWGUSI)
     /* disable interrupt */
 # ifdef HAVE_SIGPROCMASK
     sigfillset(&arg.mask);
@@ -508,7 +543,7 @@ f_trap(argc, argv)
     arg.mask = sigblock(~0);
 # endif
 
-    return rb_ensure(trap, &arg, trap_ensure, &arg);
+    return rb_ensure(trap, (VALUE)&arg, trap_ensure, (VALUE)&arg);
 #else
     return trap(&arg);
 #endif
@@ -517,8 +552,7 @@ f_trap(argc, argv)
 void
 Init_signal()
 {
-    extern VALUE mKernel;
-
+#ifndef MACOS_UNUSE_SIGNAL
     rb_define_global_function("trap", f_trap, -1);
 #ifdef POSIX_SIGNAL
     posix_signal(SIGINT, sighandle);
@@ -531,4 +565,5 @@ Init_signal()
 #ifdef SIGSEGV
     signal(SIGSEGV, sigsegv);
 #endif
+#endif /* MACOS_UNUSE_SIGNAL */
 }
